#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Header and trailer records of SEVIRI native format."""

import numpy as np

from satpy.readers.core.eum import time_cds, time_cds_expanded, time_cds_short
from satpy.readers.core.seviri import HRV_NUM_COLUMNS, HRV_NUM_LINES, VISIR_NUM_COLUMNS, VISIR_NUM_LINES


class GSDTRecords(object):
    """MSG Ground Segment Data Type records.

    Reference Document (EUM/MSG/SPE/055):
    MSG Ground Segment Design Specification (GSDS)

    """

    gp_fac_env = np.uint8
    gp_fac_id = np.uint8
    gp_sc_id = np.uint16
    gp_su_id = np.uint32
    gp_svce_type = np.uint8

    # 4 bytes
    gp_cpu_address = [
        ("Qualifier_1", np.uint8),
        ("Qualifier_2", np.uint8),
        ("Qualifier_3", np.uint8),
        ("Qualifier_4", np.uint8)
    ]

    # 22 bytes
    gp_pk_header = [
        ("HeaderVersionNo", np.uint8),
        ("PacketType", np.uint8),
        ("SubHeaderType", np.uint8),
        ("SourceFacilityId", gp_fac_id),
        ("SourceEnvId", gp_fac_env),
        ("SourceInstanceId", np.uint8),
        ("SourceSUId", gp_su_id),
        ("SourceCPUId", gp_cpu_address),
        ("DestFacilityId", gp_fac_id),
        ("DestEnvId", gp_fac_env),
        ("SequenceCount", np.uint16),
        ("PacketLength", np.int32)
    ]

    # 16 bytes
    gp_pk_sh1 = [
        ("SubHeaderVersionNo", np.uint8),
        ("ChecksumFlag", bool),
        ("Acknowledgement", (np.uint8, 4)),
        ("ServiceType", gp_svce_type),
        ("ServiceSubtype", np.uint8),
        ("PacketTime", time_cds_short),
        ("SpacecraftId", gp_sc_id)
    ]


class Msg15NativeHeaderRecord(object):
    """SEVIRI Level 1.5 header for native-format."""

    def get(self, with_archive_header):
        """Get the header type."""
        # 450400 bytes including archive header
        # 445286 bytes excluding archive header
        record = []
        if with_archive_header:
            record += [
                ("15_MAIN_PRODUCT_HEADER", L15MainProductHeaderRecord().get()),
                ("15_SECONDARY_PRODUCT_HEADER",
                 L15SecondaryProductHeaderRecord().get()),
            ]
        record += [
            ("GP_PK_HEADER", GSDTRecords.gp_pk_header),
            ("GP_PK_SH1", GSDTRecords.gp_pk_sh1),
            ("15_DATA_HEADER", L15DataHeaderRecord().get())
        ]

        return np.dtype(record).newbyteorder(">")


class L15PhData(object):
    """L15 Ph handler."""

    # 80 bytes
    l15_ph_data = [
        ("Name", "S30"),
        ("Value", "S50")
    ]


class L15MainProductHeaderRecord(object):
    """L15 Main Product header handler.

    Reference Document:
    MSG Level 1.5 Native Format File Definition
    """

    def get(self):
        """Get header data."""
        l15_ph_data = L15PhData.l15_ph_data

        l15_ph_data_identification = [
            ("Name", "S30"),
            ("Size", "S16"),
            ("Address", "S16")
        ]

        # 3674 bytes
        record = [
            ("FormatName", l15_ph_data),
            ("FormatDocumentName", l15_ph_data),
            ("FormatDocumentMajorVersion", l15_ph_data),
            ("FormatDocumentMinorVersion", l15_ph_data),
            ("CreationDateTime", l15_ph_data),
            ("CreatingCentre", l15_ph_data),
            ("DataSetIdentification", (l15_ph_data_identification, 27)),
            ("TotalFileSize", l15_ph_data),
            ("GORT", l15_ph_data),
            ("ASTI", l15_ph_data),
            ("LLOS", l15_ph_data),
            ("SNIT", l15_ph_data),
            ("AIID", l15_ph_data),
            ("SSBT", l15_ph_data),
            ("SSST", l15_ph_data),
            ("RRCC", l15_ph_data),
            ("RRBT", l15_ph_data),
            ("RRST", l15_ph_data),
            ("PPRC", l15_ph_data),
            ("PPDT", l15_ph_data),
            ("GPLV", l15_ph_data),
            ("APNM", l15_ph_data),
            ("AARF", l15_ph_data),
            ("UUDT", l15_ph_data),
            ("QQOV", l15_ph_data),
            ("UDSP", l15_ph_data)
        ]

        return record


class L15SecondaryProductHeaderRecord(object):
    """L15 Secondary Product header handler.

    Reference Document:
    MSG Level 1.5 Native Format File Definition
    """

    def get(self):
        """Get header data."""
        l15_ph_data = L15PhData.l15_ph_data

        # 1440 bytes
        record = [
            ("ABID", l15_ph_data),
            ("SMOD", l15_ph_data),
            ("APXS", l15_ph_data),
            ("AVPA", l15_ph_data),
            ("LSCD", l15_ph_data),
            ("LMAP", l15_ph_data),
            ("QDLC", l15_ph_data),
            ("QDLP", l15_ph_data),
            ("QQAI", l15_ph_data),
            ("SelectedBandIDs", l15_ph_data),
            ("SouthLineSelectedRectangle", l15_ph_data),
            ("NorthLineSelectedRectangle", l15_ph_data),
            ("EastColumnSelectedRectangle", l15_ph_data),
            ("WestColumnSelectedRectangle", l15_ph_data),
            ("NumberLinesVISIR", l15_ph_data),
            ("NumberColumnsVISIR", l15_ph_data),
            ("NumberLinesHRV", l15_ph_data),
            ("NumberColumnsHRV", l15_ph_data)
        ]

        return record


class L15DataHeaderRecord(object):
    """L15 Data Header handler.

    Reference Document (EUM/MSG/ICD/105):
    MSG Level 1.5 Image Data Format Description
    """

    def get(self):
        """Get header record data."""
        # 445248 bytes
        record = [
            ("15HeaderVersion", np.uint8),
            ("SatelliteStatus", self.satellite_status),
            ("ImageAcquisition", self.image_acquisition),
            ("CelestialEvents", self.celestial_events),
            ("ImageDescription", self.image_description),
            ("RadiometricProcessing", self.radiometric_processing),
            ("GeometricProcessing", self.geometric_processing),
            ("IMPFConfiguration", self.impf_configuration)]

        return record

    @property
    def satellite_status(self):
        """Get satellite status data."""
        # 7 bytes
        satellite_definition = [
            ("SatelliteId", np.uint16),
            ("NominalLongitude", np.float32),
            ("SatelliteStatus", np.uint8)]

        # 28 bytes
        satellite_operations = [
            ("LastManoeuvreFlag", bool),
            ("LastManoeuvreStartTime", time_cds_short),
            ("LastManoeuvreEndTime", time_cds_short),
            ("LastManoeuvreType", np.uint8),
            ("NextManoeuvreFlag", bool),
            ("NextManoeuvreStartTime", time_cds_short),
            ("NextManoeuvreEndTime", time_cds_short),
            ("NextManoeuvreType", np.uint8)]

        # 396 bytes
        orbit_coeff = [
            ("StartTime", time_cds_short),
            ("EndTime", time_cds_short),
            ("X", (np.float64, 8)),
            ("Y", (np.float64, 8)),
            ("Z", (np.float64, 8)),
            ("VX", (np.float64, 8)),
            ("VY", (np.float64, 8)),
            ("VZ", (np.float64, 8))]

        # 39612 bytes
        orbit = [
            ("PeriodStartTime", time_cds_short),
            ("PeriodEndTime", time_cds_short),
            ("OrbitPolynomial", (orbit_coeff, 100))]

        # 204 bytes
        attitude_coeff = [
            ("StartTime", time_cds_short),
            ("EndTime", time_cds_short),
            ("XofSpinAxis", (np.float64, 8)),
            ("YofSpinAxis", (np.float64, 8)),
            ("ZofSpinAxis", (np.float64, 8))]

        # 20420 bytes
        attitude = [
            ("PeriodStartTime", time_cds_short),
            ("PeriodEndTime", time_cds_short),
            ("PrincipleAxisOffsetAngle", np.float64),
            ("AttitudePolynomial", (attitude_coeff, 100))]

        # 59 bytes
        utc_correlation = [
            ("PeriodStartTime", time_cds_short),
            ("PeriodEndTime", time_cds_short),
            ("OnBoardTimeStart", (np.uint8, 7)),
            ("VarOnBoardTimeStart", np.float64),
            ("A1", np.float64),
            ("VarA1", np.float64),
            ("A2", np.float64),
            ("VarA2", np.float64)]

        # 60134 bytes
        record = [
            ("SatelliteDefinition", satellite_definition),
            ("SatelliteOperations", satellite_operations),
            ("Orbit", orbit),
            ("Attitude", attitude),
            ("SpinRetreatRCStart", np.float64),
            ("UTCCorrelation", utc_correlation)]

        return record

    @property
    def image_acquisition(self):
        """Get image acquisition data."""
        planned_acquisition_time = [
            ("TrueRepeatCycleStart", time_cds_expanded),
            ("PlanForwardScanEnd", time_cds_expanded),
            ("PlannedRepeatCycleEnd", time_cds_expanded)]

        radiometer_status = [
            ("ChannelStatus", (np.uint8, 12)),
            ("DetectorStatus", (np.uint8, 42))]

        hrv_frame_offsets = [
            ("MDUNomHRVDelay1", np.uint16),
            ("MDUNomHRVDelay2", np.uint16),
            ("Spare", np.uint16),
            ("MDUNomHRVBreakLine", np.uint16)]

        operation_parameters = [
            ("L0_LineCounter", np.uint16),
            ("K1_RetraceLines", np.uint16),
            ("K2_PauseDeciseconds", np.uint16),
            ("K3_RetraceLines", np.uint16),
            ("K4_PauseDeciseconds", np.uint16),
            ("K5_RetraceLines", np.uint16),
            ("XDeepSpaceWindowPosition", np.uint8)]

        radiometer_settings = [
            ("MDUSamplingDelays", (np.uint16, 42)),
            ("HRVFrameOffsets", hrv_frame_offsets),
            ("DHSSSynchSelection", np.uint8),
            ("MDUOutGain", (np.uint16, 42)),
            ("MDUCoarseGain", (np.uint8, 42)),
            ("MDUFineGain", (np.uint16, 42)),
            ("MDUNumericalOffset", (np.uint16, 42)),
            ("PUGain", (np.uint16, 42)),
            ("PUOffset", (np.uint16, 27)),
            ("PUBias", (np.uint16, 15)),
            ("OperationParameters", operation_parameters),
            ("RefocusingLines", np.uint16),
            ("RefocusingDirection", np.uint8),
            ("RefocusingPosition", np.uint16),
            ("ScanRefPosFlag", bool),
            ("ScanRefPosNumber", np.uint16),
            ("ScanRefPosVal", np.float32),
            ("ScanFirstLine", np.uint16),
            ("ScanLastLine", np.uint16),
            ("RetraceStartLine", np.uint16)]

        decontamination = [
            ("DecontaminationNow", bool),
            ("DecontaminationStart", time_cds_short),
            ("DecontaminationEnd", time_cds_short)]

        radiometer_operations = [
            ("LastGainChangeFlag", bool),
            ("LastGainChangeTime", time_cds_short),
            ("Decontamination", decontamination),
            ("BBCalScheduled", bool),
            ("BBCalibrationType", np.uint8),
            ("BBFirstLine", np.uint16),
            ("BBLastLine", np.uint16),
            ("ColdFocalPlaneOpTemp", np.uint16),
            ("WarmFocalPlaneOpTemp", np.uint16)]

        record = [
            ("PlannedAcquisitionTime", planned_acquisition_time),
            ("RadiometerStatus", radiometer_status),
            ("RadiometerSettings", radiometer_settings),
            ("RadiometerOperations", radiometer_operations)]

        return record

    @property
    def celestial_events(self):
        """Get celestial events data."""
        earth_moon_sun_coeff = [
            ("StartTime", time_cds_short),
            ("EndTime", time_cds_short),
            ("AlphaCoef", (np.float64, 8)),
            ("BetaCoef", (np.float64, 8))]

        star_coeff = [
            ("StarId", np.uint16),
            ("StartTime", time_cds_short),
            ("EndTime", time_cds_short),
            ("AlphaCoef", (np.float64, 8)),
            ("BetaCoef", (np.float64, 8))]

        ephemeris = [
            ("PeriodTimeStart", time_cds_short),
            ("PeriodTimeEnd", time_cds_short),
            ("RelatedOrbitFileTime", "S15"),
            ("RelatedAttitudeFileTime", "S15"),
            ("EarthEphemeris", (earth_moon_sun_coeff, 100)),
            ("MoonEphemeris", (earth_moon_sun_coeff, 100)),
            ("SunEphemeris", (earth_moon_sun_coeff, 100)),
            ("StarEphemeris", (star_coeff, (20, 100)))]

        relation_to_image = [
            ("TypeOfEclipse", np.uint8),
            ("EclipseStartTime", time_cds_short),
            ("EclipseEndTime", time_cds_short),
            ("VisibleBodiesInImage", np.uint8),
            ("BodiesCloseToFOV", np.uint8),
            ("ImpactOnImageQuality", np.uint8)]

        record = [
            ("CelestialBodiesPosition", ephemeris),
            ("RelationToImage", relation_to_image)]

        return record

    @property
    def image_description(self):
        """Get image description data."""
        projection_description = [
            ("TypeOfProjection", np.uint8),
            ("LongitudeOfSSP", np.float32)]

        reference_grid = [
            ("NumberOfLines", np.int32),
            ("NumberOfColumns", np.int32),
            ("LineDirGridStep", np.float32),
            ("ColumnDirGridStep", np.float32),
            ("GridOrigin", np.uint8)]

        planned_coverage_vis_ir = [
            ("SouthernLinePlanned", np.int32),
            ("NorthernLinePlanned", np.int32),
            ("EasternColumnPlanned", np.int32),
            ("WesternColumnPlanned", np.int32)]

        planned_coverage_hrv = [
            ("LowerSouthLinePlanned", np.int32),
            ("LowerNorthLinePlanned", np.int32),
            ("LowerEastColumnPlanned", np.int32),
            ("LowerWestColumnPlanned", np.int32),
            ("UpperSouthLinePlanned", np.int32),
            ("UpperNorthLinePlanned", np.int32),
            ("UpperEastColumnPlanned", np.int32),
            ("UpperWestColumnPlanned", np.int32)]

        level_15_image_production = [
            ("ImageProcDirection", np.uint8),
            ("PixelGenDirection", np.uint8),
            ("PlannedChanProcessing", (np.uint8, 12))]

        record = [
            ("ProjectionDescription", projection_description),
            ("ReferenceGridVIS_IR", reference_grid),
            ("ReferenceGridHRV", reference_grid),
            ("PlannedCoverageVIS_IR", planned_coverage_vis_ir),
            ("PlannedCoverageHRV", planned_coverage_hrv),
            ("Level15ImageProduction", level_15_image_production)]

        return record

    @property
    def radiometric_processing(self):
        """Get radiometric processing data."""
        rp_summary = [
            ("RadianceLinearization", (bool, 12)),
            ("DetectorEqualization", (bool, 12)),
            ("OnboardCalibrationResult", (bool, 12)),
            ("MPEFCalFeedback", (bool, 12)),
            ("MTFAdaptation", (bool, 12)),
            ("StrayLightCorrection", (bool, 12))]

        level_15_image_calibration = [
            ("CalSlope", np.float64),
            ("CalOffset", np.float64)]

        time_cuc_size = [
            ("CT1", np.uint8),
            ("CT2", np.uint8),
            ("CT3", np.uint8),
            ("CT4", np.uint8),
            ("FT1", np.uint8),
            ("FT2", np.uint8),
            ("FT3", np.uint8)]

        cold_fp_temperature = [
            ("FCUNominalColdFocalPlaneTemp", np.uint16),
            ("FCURedundantColdFocalPlaneTemp", np.uint16)]

        warm_fp_temperature = [
            ("FCUNominalWarmFocalPlaneVHROTemp", np.uint16),
            ("FCURedundantWarmFocalPlaneVHROTemp", np.uint16)]

        scan_mirror_temperature = [
            ("FCUNominalScanMirrorSensor1Temp", np.uint16),
            ("FCURedundantScanMirrorSensor1Temp", np.uint16),
            ("FCUNominalScanMirrorSensor2Temp", np.uint16),
            ("FCURedundantScanMirrorSensor2Temp", np.uint16)]

        m1m2m3_temperature = [
            ("FCUNominalM1MirrorSensor1Temp", np.uint16),
            ("FCURedundantM1MirrorSensor1Temp", np.uint16),
            ("FCUNominalM1MirrorSensor2Temp", np.uint16),
            ("FCURedundantM1MirrorSensor2Temp", np.uint16),
            ("FCUNominalM23AssemblySensor1Temp", np.uint8),
            ("FCURedundantM23AssemblySensor1Temp", np.uint8),
            ("FCUNominalM23AssemblySensor2Temp", np.uint8),
            ("FCURedundantM23AssemblySensor2Temp", np.uint8)]

        baffle_temperature = [
            ("FCUNominalM1BaffleTemp", np.uint16),
            ("FCURedundantM1BaffleTemp", np.uint16)]

        blackbody_temperature = [
            ("FCUNominalBlackBodySensorTemp", np.uint16),
            ("FCURedundantBlackBodySensorTemp", np.uint16)]

        fcu_mode = [
            ("FCUNominalSMMStatus", "S2"),
            ("FCURedundantSMMStatus", "S2")]

        extracted_bb_data = [
            ("NumberOfPixelsUsed", np.uint32),
            ("MeanCount", np.float32),
            ("RMS", np.float32),
            ("MaxCount", np.uint16),
            ("MinCount", np.uint16),
            ("BB_Processing_Slope", np.float64),
            ("BB_Processing_Offset", np.float64)]

        bb_related_data = [
            ("OnBoardBBTime", time_cuc_size),
            ("MDUOutGain", (np.uint16, 42)),
            ("MDUCoarseGain", (np.uint8, 42)),
            ("MDUFineGain", (np.uint16, 42)),
            ("MDUNumericalOffset", (np.uint16, 42)),
            ("PUGain", (np.uint16, 42)),
            ("PUOffset", (np.uint16, 27)),
            ("PUBias", (np.uint16, 15)),
            ("DCRValues", (np.uint8, 63)),
            ("X_DeepSpaceWindowPosition", np.int8),
            ("ColdFPTemperature", cold_fp_temperature),
            ("WarmFPTemperature", warm_fp_temperature),
            ("ScanMirrorTemperature", scan_mirror_temperature),
            ("M1M2M3Temperature", m1m2m3_temperature),
            ("BaffleTemperature", baffle_temperature),
            ("BlackBodyTemperature", blackbody_temperature),
            ("FCUMode", fcu_mode),
            ("ExtractedBBData", (extracted_bb_data, 12))]

        black_body_data_used = [
            ("BBObservationUTC", time_cds_expanded),
            ("BBRelatedData", bb_related_data)]

        impf_cal_data = [
            ("ImageQualityFlag", np.uint8),
            ("ReferenceDataFlag", np.uint8),
            ("AbsCalMethod", np.uint8),
            ("Pad1", "S1"),
            ("AbsCalWeightVic", np.float32),
            ("AbsCalWeightXsat", np.float32),
            ("AbsCalCoeff", np.float32),
            ("AbsCalError", np.float32),
            ("GSICSCalCoeff", np.float32),
            ("GSICSCalError", np.float32),
            ("GSICSOffsetCount", np.float32)]

        rad_proc_mtf_adaptation = [
            ("VIS_IRMTFCorrectionE_W", (np.float32, (33, 16))),
            ("VIS_IRMTFCorrectionN_S", (np.float32, (33, 16))),
            ("HRVMTFCorrectionE_W", (np.float32, (9, 16))),
            ("HRVMTFCorrectionN_S", (np.float32, (9, 16))),
            ("StraylightCorrection", (np.float32, (12, 8, 8)))]

        record = [
            ("RPSummary", rp_summary),
            ("Level15ImageCalibration", (level_15_image_calibration, 12)),
            ("BlackBodyDataUsed", black_body_data_used),
            ("MPEFCalFeedback", (impf_cal_data, 12)),
            ("RadTransform", (np.float32, (42, 64))),
            ("RadProcMTFAdaptation", rad_proc_mtf_adaptation)]

        return record

    @property
    def geometric_processing(self):
        """Get geometric processing data."""
        opt_axis_distances = [
            ("E-WFocalPlane", (np.float32, 42)),
            ("N_SFocalPlane", (np.float32, 42))]

        earth_model = [
            ("TypeOfEarthModel", np.uint8),
            ("EquatorialRadius", np.float64),
            ("NorthPolarRadius", np.float64),
            ("SouthPolarRadius", np.float64)]

        record = [
            ("OptAxisDistances", opt_axis_distances),
            ("EarthModel", earth_model),
            ("AtmosphericModel", (np.float32, (12, 360))),
            ("ResamplingFunctions", (np.uint8, 12))]

        return record

    @property
    def impf_configuration(self):
        """Get impf configuration information."""
        overall_configuration = [
            ("Issue", np.uint16),
            ("Revision", np.uint16)
        ]

        sw_version = overall_configuration

        info_base_versions = sw_version

        su_configuration = [
            ("SWVersion", sw_version),
            ("InfoBaseVersions", (info_base_versions, 10))
        ]

        su_details = [
            ("SUId", GSDTRecords.gp_su_id),
            ("SUIdInstance", np.int8),
            ("SUMode", np.uint8),
            ("SUState", np.uint8),
            ("SUConfiguration", su_configuration)
        ]

        equalisation_params = [
            ("ConstCoeff", np.float32),
            ("LinearCoeff", np.float32),
            ("QuadraticCoeff", np.float32)
        ]

        black_body_data_for_warm_start = [
            ("GTotalForMethod1", (np.float64, 12)),
            ("GTotalForMethod2", (np.float64, 12)),
            ("GTotalForMethod3", (np.float64, 12)),
            ("GBackForMethod1", (np.float64, 12)),
            ("GBackForMethod2", (np.float64, 12)),
            ("GBackForMethod3", (np.float64, 12)),
            ("RatioGTotalToGBack", (np.float64, 12)),
            ("GainInFrontOpticsCont", (np.float64, 12)),
            ("CalibrationConstants", (np.float32, 12)),
            ("maxIncidentRadiance", (np.float64, 12)),
            ("TimeOfColdObsSeconds", np.float64),
            ("TimeOfColdObsNanoSecs", np.float64),
            ("IncidenceRadiance", (np.float64, 12)),
            ("TempCal", np.float64),
            ("TempM1", np.float64),
            ("TempScan", np.float64),
            ("TempM1Baf", np.float64),
            ("TempCalSurround", np.float64)
        ]

        mirror_parameters = [
            ("MaxFeedbackVoltage", np.float64),
            ("MinFeedbackVoltage", np.float64),
            ("MirrorSlipEstimate", np.float64)
        ]

        hktm_parameters = [
            ("TimeS0Packet", time_cds_short),
            ("TimeS1Packet", time_cds_short),
            ("TimeS2Packet", time_cds_short),
            ("TimeS3Packet", time_cds_short),
            ("TimeS4Packet", time_cds_short),
            ("TimeS5Packet", time_cds_short),
            ("TimeS6Packet", time_cds_short),
            ("TimeS7Packet", time_cds_short),
            ("TimeS8Packet", time_cds_short),
            ("TimeS9Packet", time_cds_short),
            ("TimeSYPacket", time_cds_short),
            ("TimePSPacket", time_cds_short)
        ]

        warm_start_params = [
            ("ScanningLaw", (np.float64, 1527)),
            ("RadFramesAlignment", (np.float64, 3)),
            ("ScanningLawVariation", (np.float32, 2)),
            ("EqualisationParams", (equalisation_params, 42)),
            ("BlackBodyDataForWarmStart", black_body_data_for_warm_start),
            ("MirrorParameters", mirror_parameters),
            ("LastSpinPeriod", np.float64),
            ("HKTMParameters", hktm_parameters),
            ("WSPReserved", (np.uint8, 3312))
        ]

        record = [
            ("OverallConfiguration", overall_configuration),
            ("SUDetails", (su_details, 50)),
            ("WarmStartParams", warm_start_params)
        ]

        return record


class Msg15NativeTrailerRecord(object):
    """SEVIRI Level 1.5 trailer for native-format.

    Reference Document (EUM/MSG/ICD/105):
    MSG Level 1.5 Image Data Format Description
    """

    def get(self):
        """Get header record data."""
        # 380363 bytes
        record = [
            ("GP_PK_HEADER", GSDTRecords.gp_pk_header),
            ("GP_PK_SH1", GSDTRecords.gp_pk_sh1),
            ("15TRAILER", self.seviri_l15_trailer)
        ]

        return np.dtype(record).newbyteorder(">")

    @property
    def seviri_l15_trailer(self):
        """Get file trailer data."""
        record = [
            ("15TrailerVersion", np.uint8),
            ("ImageProductionStats", self.image_production_stats),
            ("NavigationExtractionResults", self.navigation_extraction_results),
            ("RadiometricQuality", self.radiometric_quality),
            ("GeometricQuality", self.geometric_quality),
            ("TimelinessAndCompleteness", self.timeliness_and_completeness)
        ]
        return record

    @property
    def image_production_stats(self):
        """Get image production statistics."""
        gp_sc_id = GSDTRecords.gp_sc_id

        actual_scanning_summary = [
            ("NominalImageScanning", np.uint8),
            ("ReducedScan", np.uint8),
            ("ForwardScanStart", time_cds_short),
            ("ForwardScanEnd", time_cds_short)
        ]

        radiometric_behaviour = [
            ("NominalBehaviour", np.uint8),
            ("RadScanIrregularity", np.uint8),
            ("RadStoppage", np.uint8),
            ("RepeatCycleNotCompleted", np.uint8),
            ("GainChangeTookPlace", np.uint8),
            ("DecontaminationTookPlace", np.uint8),
            ("NoBBCalibrationAchieved", np.uint8),
            ("IncorrectTemperature", np.uint8),
            ("InvalidBBData", np.uint8),
            ("InvalidAuxOrHKTMData", np.uint8),
            ("RefocusingMechanismActuated", np.uint8),
            ("MirrorBackToReferencePos", np.uint8)
        ]

        reception_summary_stats = [
            ("PlannedNumberOfL10Lines", (np.uint32, 12)),
            ("NumberOfMissingL10Lines", (np.uint32, 12)),
            ("NumberOfCorruptedL10Lines", (np.uint32, 12)),
            ("NumberOfReplacedL10Lines", (np.uint32, 12))
        ]

        l15_image_validity = [
            ("NominalImage", np.uint8),
            ("NonNominalBecauseIncomplete", np.uint8),
            ("NonNominalRadiometricQuality", np.uint8),
            ("NonNominalGeometricQuality", np.uint8),
            ("NonNominalTimeliness", np.uint8),
            ("IncompleteL15", np.uint8),
        ]

        actual_l15_coverage_vis_ir = [
            ("SouthernLineActual", np.int32),
            ("NorthernLineActual", np.int32),
            ("EasternColumnActual", np.int32),
            ("WesternColumnActual", np.int32)
        ]

        actual_l15_coverage_hrv = [
            ("LowerSouthLineActual", np.int32),
            ("LowerNorthLineActual", np.int32),
            ("LowerEastColumnActual", np.int32),
            ("LowerWestColumnActual", np.int32),
            ("UpperSouthLineActual", np.int32),
            ("UpperNorthLineActual", np.int32),
            ("UpperEastColumnActual", np.int32),
            ("UpperWestColumnActual", np.int32),
        ]

        record = [
            ("SatelliteId", gp_sc_id),
            ("ActualScanningSummary", actual_scanning_summary),
            ("RadiometricBehaviour", radiometric_behaviour),
            ("ReceptionSummaryStats", reception_summary_stats),
            ("L15ImageValidity", (l15_image_validity, 12)),
            ("ActualL15CoverageVIS_IR", actual_l15_coverage_vis_ir),
            ("ActualL15CoverageHRV", actual_l15_coverage_hrv)
        ]

        return record

    @property
    def navigation_extraction_results(self):
        """Get navigation extraction data."""
        horizon_observation = [
            ("HorizonId", np.uint8),
            ("Alpha", np.float64),
            ("AlphaConfidence", np.float64),
            ("Beta", np.float64),
            ("BetaConfidence", np.float64),
            ("ObservationTime", time_cds),
            ("SpinRate", np.float64),
            ("AlphaDeviation", np.float64),
            ("BetaDeviation", np.float64)
        ]

        star_observation = [
            ("StarId", np.uint16),
            ("Alpha", np.float64),
            ("AlphaConfidence", np.float64),
            ("Beta", np.float64),
            ("BetaConfidence", np.float64),
            ("ObservationTime", time_cds),
            ("SpinRate", np.float64),
            ("AlphaDeviation", np.float64),
            ("BetaDeviation", np.float64)
        ]

        landmark_observation = [
            ("LandmarkId", np.uint16),
            ("LandmarkLongitude", np.float64),
            ("LandmarkLatitude", np.float64),
            ("Alpha", np.float64),
            ("AlphaConfidence", np.float64),
            ("Beta", np.float64),
            ("BetaConfidence", np.float64),
            ("ObservationTime", time_cds),
            ("SpinRate", np.float64),
            ("AlphaDeviation", np.float64),
            ("BetaDeviation", np.float64)
        ]

        record = [
            ("ExtractedHorizons", (horizon_observation, 4)),
            ("ExtractedStars", (star_observation, 20)),
            ("ExtractedLandmarks", (landmark_observation, 50))
        ]

        return record

    @property
    def radiometric_quality(self):
        """Get radiometric quality record data."""
        l10_rad_quality = [
            ("FullImageMinimumCount", np.uint16),
            ("FullImageMaximumCount", np.uint16),
            ("EarthDiskMinimumCount", np.uint16),
            ("EarthDiskMaximumCount", np.uint16),
            ("MoonMinimumCount", np.uint16),
            ("MoonMaximumCount", np.uint16),
            ("FullImageMeanCount", np.float32),
            ("FullImageStandardDeviation", np.float32),
            ("EarthDiskMeanCount", np.float32),
            ("EarthDiskStandardDeviation", np.float32),
            ("MoonMeanCount", np.float32),
            ("MoonStandardDeviation", np.float32),
            ("SpaceMeanCount", np.float32),
            ("SpaceStandardDeviation", np.float32),
            ("SESpaceCornerMeanCount", np.float32),
            ("SESpaceCornerStandardDeviation", np.float32),
            ("SWSpaceCornerMeanCount", np.float32),
            ("SWSpaceCornerStandardDeviation", np.float32),
            ("NESpaceCornerMeanCount", np.float32),
            ("NESpaceCornerStandardDeviation", np.float32),
            ("NWSpaceCornerMeanCount", np.float32),
            ("NWSpaceCornerStandardDeviation", np.float32),
            ("4SpaceCornersMeanCount", np.float32),
            ("4SpaceCornersStandardDeviation", np.float32),
            ("FullImageHistogram", (np.uint32, 256)),
            ("EarthDiskHistogram", (np.uint32, 256)),
            ("ImageCentreSquareHistogram", (np.uint32, 256)),
            ("SESpaceCornerHistogram", (np.uint32, 128)),
            ("SWSpaceCornerHistogram", (np.uint32, 128)),
            ("NESpaceCornerHistogram", (np.uint32, 128)),
            ("NWSpaceCornerHistogram", (np.uint32, 128)),
            ("FullImageEntropy", (np.float32, 3)),
            ("EarthDiskEntropy", (np.float32, 3)),
            ("ImageCentreSquareEntropy", (np.float32, 3)),
            ("SESpaceCornerEntropy", (np.float32, 3)),
            ("SWSpaceCornerEntropy", (np.float32, 3)),
            ("NESpaceCornerEntropy", (np.float32, 3)),
            ("NWSpaceCornerEntropy", (np.float32, 3)),
            ("4SpaceCornersEntropy", (np.float32, 3)),
            ("ImageCentreSquarePSD_EW", (np.float32, 128)),
            ("FullImagePSD_EW", (np.float32, 128)),
            ("ImageCentreSquarePSD_NS", (np.float32, 128)),
            ("FullImagePSD_NS", (np.float32, 128))
        ]

        l15_rad_quality = [
            ("FullImageMinimumCount", np.uint16),
            ("FullImageMaximumCount", np.uint16),
            ("EarthDiskMinimumCount", np.uint16),
            ("EarthDiskMaximumCount", np.uint16),
            ("FullImageMeanCount", np.float32),
            ("FullImageStandardDeviation", np.float32),
            ("EarthDiskMeanCount", np.float32),
            ("EarthDiskStandardDeviation", np.float32),
            ("SpaceMeanCount", np.float32),
            ("SpaceStandardDeviation", np.float32),
            ("FullImageHistogram", (np.uint32, 256)),
            ("EarthDiskHistogram", (np.uint32, 256)),
            ("ImageCentreSquareHistogram", (np.uint32, 256)),
            ("FullImageEntropy", (np.float32, 3)),
            ("EarthDiskEntropy", (np.float32, 3)),
            ("ImageCentreSquareEntropy", (np.float32, 3)),
            ("ImageCentreSquarePSD_EW", (np.float32, 128)),
            ("FullImagePSD_EW", (np.float32, 128)),
            ("ImageCentreSquarePSD_NS", (np.float32, 128)),
            ("FullImagePSD_NS", (np.float32, 128)),
            ("SESpaceCornerL15_RMS", np.float32),
            ("SESpaceCornerL15_Mean", np.float32),
            ("SWSpaceCornerL15_RMS", np.float32),
            ("SWSpaceCornerL15_Mean", np.float32),
            ("NESpaceCornerL15_RMS", np.float32),
            ("NESpaceCornerL15_Mean", np.float32),
            ("NWSpaceCornerL15_RMS", np.float32),
            ("NWSpaceCornerL15_Mean", np.float32)
        ]

        record = [
            ("L10RadQuality", (l10_rad_quality, 42)),
            ("L15RadQuality", (l15_rad_quality, 12))
        ]

        return record

    @property
    def geometric_quality(self):
        """Get geometric quality record data."""
        absolute_accuracy = [
            ("QualityInfoValidity", np.uint8),
            ("EastWestAccuracyRMS", np.float32),
            ("NorthSouthAccuracyRMS", np.float32),
            ("MagnitudeRMS", np.float32),
            ("EastWestUncertaintyRMS", np.float32),
            ("NorthSouthUncertaintyRMS", np.float32),
            ("MagnitudeUncertaintyRMS", np.float32),
            ("EastWestMaxDeviation", np.float32),
            ("NorthSouthMaxDeviation", np.float32),
            ("MagnitudeMaxDeviation", np.float32),
            ("EastWestUncertaintyMax", np.float32),
            ("NorthSouthUncertaintyMax", np.float32),
            ("MagnitudeUncertaintyMax", np.float32)
        ]

        relative_accuracy = absolute_accuracy
        pixels_500_relative_accuracy = absolute_accuracy
        pixels_16_relative_accuracy = absolute_accuracy

        misregistration_residuals = [
            ("QualityInfoValidity", np.uint8),
            ("EastWestResidual", np.float32),
            ("NorthSouthResidual", np.float32),
            ("EastWestUncertainty", np.float32),
            ("NorthSouthUncertainty", np.float32),
            ("EastWestRMS", np.float32),
            ("NorthSouthRMS", np.float32),
            ("EastWestMagnitude", np.float32),
            ("NorthSouthMagnitude", np.float32),
            ("EastWestMagnitudeUncertainty", np.float32),
            ("NorthSouthMagnitudeUncertainty", np.float32)
        ]

        geometric_quality_status = [
            ("QualityNominal", np.uint8),
            ("NominalAbsolute", np.uint8),
            ("NominalRelativeToPreviousImage", np.uint8),
            ("NominalForREL500", np.uint8),
            ("NominalForREL16", np.uint8),
            ("NominalForResMisreg", np.uint8)
        ]

        record = [
            ("AbsoluteAccuracy", (absolute_accuracy, 12)),
            ("RelativeAccuracy", (relative_accuracy, 12)),
            ("500PixelsRelativeAccuracy", (pixels_500_relative_accuracy, 12)),
            ("16PixelsRelativeAccuracy", (pixels_16_relative_accuracy, 12)),
            ("MisregistrationResiduals", (misregistration_residuals, 12)),
            ("GeometricQualityStatus", (geometric_quality_status, 12))
        ]

        return record

    @property
    def timeliness_and_completeness(self):
        """Get time and completeness record data."""
        timeliness = [
            ("MaxDelay", np.float32),
            ("MinDelay", np.float32),
            ("MeanDelay", np.float32)
        ]

        completeness = [
            ("PlannedL15ImageLines", np.uint16),
            ("GeneratedL15ImageLines", np.uint16),
            ("ValidL15ImageLines", np.uint16),
            ("DummyL15ImageLines", np.uint16),
            ("CorruptedL15ImageLines", np.uint16)
        ]

        record = [
            ("Timeliness", timeliness),
            ("Completeness", (completeness, 12))
        ]

        return record


class HritPrologue(L15DataHeaderRecord):
    """HRIT Prologue handler."""

    def get(self):
        """Get record data array."""
        # X bytes
        record = [
            ("SatelliteStatus", self.satellite_status),
            ("ImageAcquisition", self.image_acquisition),
            ("CelestialEvents", self.celestial_events),
            ("ImageDescription", self.image_description),
            ("RadiometricProcessing", self.radiometric_processing),
            ("GeometricProcessing", self.geometric_processing)
        ]

        return np.dtype(record).newbyteorder(">")


def get_native_header(with_archive_header=True):
    """Get Native format header type.

    There are two variants, one including an ASCII archive header and one
    without that header. The header is prepended if the data are ordered
    through the EUMETSAT data center.
    """
    return Msg15NativeHeaderRecord().get(with_archive_header)


DEFAULT_15_SECONDARY_PRODUCT_HEADER = {
    "NorthLineSelectedRectangle": {"Value": VISIR_NUM_LINES},
    "SouthLineSelectedRectangle": {"Value": 1},
    "EastColumnSelectedRectangle": {"Value": 1},
    "WestColumnSelectedRectangle": {"Value": VISIR_NUM_COLUMNS},
    "NumberColumnsVISIR": {"Value": VISIR_NUM_COLUMNS},
    "NumberLinesVISIR": {"Value": VISIR_NUM_LINES},
    "NumberColumnsHRV": {"Value": HRV_NUM_COLUMNS},
    "NumberLinesHRV": {"Value": HRV_NUM_LINES},
    "SelectedBandIDs": {"Value": "XXXXXXXXXXXX"}
}
"""Default secondary product header for files containing all channels."""


hrit_epilogue = np.dtype(
    Msg15NativeTrailerRecord().seviri_l15_trailer).newbyteorder(">")
hrit_prologue = HritPrologue().get()
impf_configuration = np.dtype(
    L15DataHeaderRecord().impf_configuration).newbyteorder(">")
native_trailer = Msg15NativeTrailerRecord().get()



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Satpy developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The abi_l2_nc reader tests package."""

import contextlib
from typing import Optional
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr


def _create_cmip_dataset(data_variable: str = "HT"):
    proj = xr.DataArray(
        [],
        attrs={
            "semi_major_axis": 1.,
            "semi_minor_axis": 1.,
            "perspective_point_height": 1.,
            "longitude_of_projection_origin": -90.,
            "sweep_angle_axis": u"x"
        }
    )
    x__ = xr.DataArray(
        da.from_array([0, 1]),
        attrs={"scale_factor": 2., "add_offset": -1.},
        dims=("x",),
    )
    y__ = xr.DataArray(
        da.from_array([0, 1]),
        attrs={"scale_factor": -2., "add_offset": 1.},
        dims=("y",),
    )

    ht_da = xr.DataArray(da.from_array(np.array([2, -1, -32768, 32767]).astype(np.int16).reshape((2, 2))),
                         dims=("y", "x"),
                         attrs={"scale_factor": 0.3052037,
                                "add_offset": 0.,
                                "_FillValue": np.array(-1).astype(np.int16),
                                "_Unsigned": "True",
                                "units": "m"},)

    fake_dataset = xr.Dataset(
        data_vars={
            "goes_imager_projection": proj,
            "x": x__,
            "y": y__,
            data_variable: ht_da,
            "nominal_satellite_subpoint_lat": np.array(0.0),
            "nominal_satellite_subpoint_lon": np.array(-89.5),
            "nominal_satellite_height": np.array(35786.02),
            "spatial_resolution": "10km at nadir",

        },
        attrs={
            "time_coverage_start": "2017-09-20T17:30:40.8Z",
            "time_coverage_end": "2017-09-20T17:41:17.5Z",
            "spatial_resolution": "2km at nadir",
        }
    )
    return fake_dataset


def _compare_subdict(actual_dict, exp_sub_dict):
    for key, value in exp_sub_dict.items():
        assert key in actual_dict
        assert actual_dict[key] == value


def _assert_orbital_parameters(orb_params):
    assert orb_params["satellite_nominal_longitude"] == -89.5
    assert orb_params["satellite_nominal_latitude"] == 0.0
    assert orb_params["satellite_nominal_altitude"] == 35786020.0


def _create_mcmip_dataset():
    ds1 = _create_cmip_dataset("CMI_C01")
    ds2 = _create_cmip_dataset("CMI_C14")
    ds1["CMI_C01"].attrs["units"] = "1"
    ds2["CMI_C14"].attrs["units"] = "K"
    ds1["CMI_C14"] = ds2["CMI_C14"]
    return ds1


def _create_aod_dataset():
    ds1 = _create_cmip_dataset("AOD")
    ds1["AOD"].attrs["units"] = "1"
    return ds1


def _create_sst_dataset():
    ds1 = _create_cmip_dataset("SST")
    ds1["SST"].attrs["units"] = "K"
    dqf_data = np.zeros_like(ds1["SST"], dtype=np.uint8)
    dqf_data[-1, -1] = 1
    dqf = xr.DataArray(da.from_array(dqf_data),
                       dims=ds1["SST"].dims,
                       attrs={
                           "_FillValue": np.uint8(255),
                           "units": "1",
                           "flag_meanings": "good_quality_qf degraded_quality_qf "
                                            "severely_degraded_quality_qf invalid_due_to_unprocessed_qf",
                           "flag_values": [np.uint8(0), np.uint8(1), np.uint8(2), np.uint8(3)],
                       })
    ds1["DQF"] = dqf
    return ds1


class Test_NC_ABI_L2_get_dataset:
    """Test get dataset function of the NC_ABI_L2 reader."""

    @pytest.mark.parametrize(
        ("obs_type", "ds_func", "var_name", "var_attrs", "fh_kwargs"),
        [
            ("ACHA", _create_cmip_dataset, "HT", {"units": "m"}, {}),
            ("AOD", _create_aod_dataset, "AOD", {"units": "1"}, {}),
            ("SST", _create_sst_dataset, "SST", {"units": "K"}, {}),
            ("SST", _create_sst_dataset, "SST", {"units": "K"}, {"filters": ["good_quality_qf"]}),
            ("SST", _create_sst_dataset, "SST", {"units": "K"}, {"filters": []}),
        ]
    )
    def test_get_dataset(self, obs_type, ds_func, var_name, var_attrs, fh_kwargs):
        """Test basic L2 load."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name=var_name)
        check_dqf = bool(fh_kwargs.get("filters"))
        with _create_reader_for_fake_data(obs_type, ds_func(), fh_kwargs=fh_kwargs) as reader:
            res = reader.get_dataset(key, {"file_key": var_name})

        exp_data = np.array([[2 * 0.3052037, np.nan],
                             [32768 * 0.3052037, 32767 * 0.3052037]])
        if check_dqf:
            exp_data[-1, -1] = np.nan

        exp_attrs = {"instrument_ID": None,
                     "modifiers": (),
                     "name": var_name,
                     "observation_type": obs_type,
                     "orbital_slot": None,
                     "platform_name": "GOES-16",
                     "platform_shortname": "G16",
                     "production_site": None,
                     "scan_mode": "M3",
                     "scene_abbr": "C",
                     "scene_id": None,
                     "sensor": "abi",
                     "timeline_ID": None,
                     }
        exp_attrs.update(var_attrs)

        np.testing.assert_allclose(res.data, exp_data, equal_nan=True)
        _compare_subdict(res.attrs, exp_attrs)
        _assert_orbital_parameters(res.attrs["orbital_parameters"])

    def test_get_dataset_gfls(self):
        """Test that Low Cloud and Fog filenames work."""
        from satpy.tests.utils import make_dataid
        filename_info = {"platform_shortname": "g16", "scene_abbr": "FD"}
        key = make_dataid(name="MVFR_Fog_Prob")
        with _create_reader_for_fake_data("GFLS", _create_cmip_dataset("MVFR_Fog_Prob"), filename_info) as reader:
            res = reader.get_dataset(key, {"file_key": "MVFR_Fog_Prob"})
        assert res.attrs["platform_name"] == "GOES-16"


class TestMCMIPReading:
    """Test cases of the MCMIP file format."""

    @pytest.mark.parametrize(
        ("product", "exp_metadata"),
        [
            ("C14", {"calibration": "brightness_temperature", "wavelength": (10.8, 11.2, 11.6), "units": "K"}),
            ("C01", {"calibration": "reflectance", "wavelength": (0.45, 0.47, 0.49), "units": "%"}),
        ]
    )
    @mock.patch("satpy.readers.core.abi.xr")
    def test_mcmip_get_dataset(self, xr_, product, exp_metadata):
        """Test getting channel from MCMIP file."""
        import datetime as dt

        from pyresample.geometry import AreaDefinition

        from satpy import Scene
        fake_ds = _create_mcmip_dataset()
        xr_.open_dataset.return_value = fake_ds

        fn = "OR_ABI-L2-MCMIPF-M6_G16_s20192600241149_e20192600243534_c20192600245360.nc"
        scn = Scene(reader="abi_l2_nc", filenames=[fn])
        scn.load([product])

        exp_data = np.array([[2 * 0.3052037, np.nan],
                             [32768 * 0.3052037, 32767 * 0.3052037]])
        if "C01" in product:
            exp_data *= 100

        exp_attrs = {
            "instrument_ID": None,
            "modifiers": (),
            "name": product,
            "observation_type": "MCMIP",
            "orbital_slot": None,
            "reader": "abi_l2_nc",
            "platform_name": "GOES-16",
            "platform_shortname": "G16",
            "production_site": None,
            "scan_mode": "M6",
            "scene_abbr": "F",
            "scene_id": None,
            "sensor": "abi",
            "timeline_ID": None,
            "start_time": dt.datetime(2017, 9, 20, 17, 30, 40, 800000),
            "end_time": dt.datetime(2017, 9, 20, 17, 41, 17, 500000),
            "ancillary_variables": [],
        }
        exp_attrs.update(exp_metadata)

        res = scn[product]
        np.testing.assert_allclose(res.data, exp_data, equal_nan=True)
        assert isinstance(res.attrs["area"], AreaDefinition)
        _compare_subdict(res.attrs, exp_attrs)
        _assert_orbital_parameters(res.attrs["orbital_parameters"])


class Test_NC_ABI_L2_area_fixedgrid:
    """Test the NC_ABI_L2 reader."""

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_fixedgrid(self, adef):
        """Test the area generation."""
        with _create_reader_for_fake_data("RSR", _create_cmip_dataset()) as reader:
            reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"a": 1.0, "b": 1.0, "h": 1.0, "lon_0": -90.0,
                                "proj": "geos", "sweep": "x", "units": "m"}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-2., -2.,  2.,  2.))


class Test_NC_ABI_L2_area_latlon:
    """Test the NC_ABI_L2 reader."""

    def setup_method(self):
        """Create fake data for the tests."""
        proj = xr.DataArray(
            [],
            attrs={"semi_major_axis": 1.,
                   "semi_minor_axis": 1.,
                   "inverse_flattening": 1.,
                   "longitude_of_prime_meridian": 0.0,
                   }
        )

        proj_ext = xr.DataArray(
            [],
            attrs={"geospatial_westbound_longitude": -85.0,
                   "geospatial_eastbound_longitude": -65.0,
                   "geospatial_northbound_latitude": 20.0,
                   "geospatial_southbound_latitude": -20.0,
                   "geospatial_lat_center": 0.0,
                   "geospatial_lon_center": -75.0,
                   })

        x__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": 2., "add_offset": -1.},
            dims=("lon",),
        )
        y__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": -2., "add_offset": 1.},
            dims=("lat",),
        )
        fake_dataset = xr.Dataset(
            data_vars={
                "goes_lat_lon_projection": proj,
                "geospatial_lat_lon_extent": proj_ext,
                "lon": x__,
                "lat": y__,
                "RSR": xr.DataArray(np.ones((2, 2)), dims=("lat", "lon")),
            },
        )
        self.fake_dataset = fake_dataset

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_latlon(self, adef):
        """Test the area generation."""
        with _create_reader_for_fake_data("RSR", self.fake_dataset) as reader:
            reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"proj": "latlong", "a": 1.0, "b": 1.0, "fi": 1.0, "pm": 0.0,
                                "lon_0": -75.0, "lat_0": 0.0}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-85.0, -20.0, -65.0, 20))


class Test_NC_ABI_L2_area_AOD:
    """Test the NC_ABI_L2 reader for the AOD product."""

    def setup_method(self, xr_):
        """Create fake data for the tests."""
        proj = xr.DataArray(
            [],
            attrs={"semi_major_axis": 1.,
                   "semi_minor_axis": 1.,
                   "inverse_flattening": 1.,
                   "longitude_of_prime_meridian": 0.0,
                   }
        )

        proj_ext = xr.DataArray(
            [],
            attrs={"geospatial_westbound_longitude": -85.0,
                   "geospatial_eastbound_longitude": -65.0,
                   "geospatial_northbound_latitude": 20.0,
                   "geospatial_southbound_latitude": -20.0,
                   "geospatial_lat_center": 0.0,
                   "geospatial_lon_center": -75.0,
                   })

        x__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": 2., "add_offset": -1.},
            dims=("x",),
        )
        y__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": -2., "add_offset": 1.},
            dims=("y",),
        )
        fake_dataset = xr.Dataset(
            data_vars={
                "goes_lat_lon_projection": proj,
                "geospatial_lat_lon_extent": proj_ext,
                "x": x__,
                "y": y__,
                "RSR": xr.DataArray(np.ones((2, 2)), dims=("y", "x")),
            },
        )
        self.fake_dataset = fake_dataset

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_xy(self, adef):
        """Test the area generation."""
        with _create_reader_for_fake_data("RSR", self.fake_dataset) as reader:
            reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"proj": "latlong", "a": 1.0, "b": 1.0, "fi": 1.0, "pm": 0.0,
                                "lon_0": -75.0, "lat_0": 0.0}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-85.0, -20.0, -65.0, 20))


@contextlib.contextmanager
def _create_reader_for_fake_data(
        observation_type: str,
        fake_dataset: xr.Dataset,
        filename_info: Optional[dict] = None,
        fh_kwargs: dict | None = None,
):
    from satpy.readers.abi_l2_nc import NC_ABI_L2

    if filename_info is None:
        filename_info = {
            "platform_shortname": "G16",
            "scene_abbr": "C", "scan_mode": "M3"
        }
    fh_args = (
        "filename",
        filename_info,
        {"file_type": "info", "observation_type": observation_type},
    )
    if fh_kwargs is None:
        fh_kwargs = {}
    fh_cls = NC_ABI_L2
    with mock.patch("satpy.readers.core.abi.xr") as xr_:
        xr_.open_dataset.return_value = fake_dataset
        file_handler = fh_cls(*fh_args, **fh_kwargs)
        yield file_handler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'mersi2_l1b' reader."""
import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler


def _get_calibration(num_scans, ftype):
    calibration = {
        f"Calibration/{ftype}_Cal_Coeff":
            xr.DataArray(
                da.ones((19, 3), chunks=1024, dtype=np.float32),
                attrs={"Slope": np.array([1.] * 19, dtype=np.float32),
                       "Intercept": np.array([0.] * 19, dtype=np.float32)},
                dims=("_bands", "_coeffs")),
        "Calibration/Solar_Irradiance":
            xr.DataArray(
                da.ones((19, ), chunks=1024, dtype=np.float32),
                attrs={"Slope": np.array([1.] * 19, dtype=np.float32),
                       "Intercept": np.array([0.] * 19, dtype=np.float32)},
                dims=("_bands")),
        "Calibration/Solar_Irradiance_LL":
            xr.DataArray(
                da.ones((1, ), chunks=1024, dtype=np.float32),
                attrs={"Slope": np.array([1.], dtype=np.float32),
                       "Intercept": np.array([0.], dtype=np.float32)},
                dims=("_bands")),
        "Calibration/IR_Cal_Coeff":
            xr.DataArray(
                da.ones((6, 4, num_scans), chunks=1024, dtype=np.float32),
                attrs={"Slope": np.array([1.] * 6, dtype=np.float32),
                       "Intercept": np.array([0.] * 6, dtype=np.float32)},
                dims=("_bands", "_coeffs", "_scans")),
    }
    return calibration


def _get_250m_data(num_scans, rows_per_scan, num_cols, filetype_info):
    # Set some default attributes
    is_fy3ab_mersi1 = filetype_info["file_type"].startswith(("fy3a_mersi1", "fy3b_mersi1"))

    fill_value_name = "_FillValue" if is_fy3ab_mersi1 else "FillValue"
    key_prefix = "" if is_fy3ab_mersi1 else "Data/"

    def_attrs = {fill_value_name: 65535,
                 "valid_range": [0, 4095],
                 "Slope": np.array([1.] * 1, dtype=np.float32), "Intercept": np.array([0.] * 1, dtype=np.float32)
                 }
    nounits_attrs = {**def_attrs, **{"units": "NO"}}
    radunits_attrs = {**def_attrs, **{"units": "mW/ (m2 cm-1 sr)"}}
    valid_range_none_attrs = radunits_attrs.copy()
    valid_range_none_attrs["valid_range"] = None

    data = {
        f"{key_prefix}EV_250_RefSB_b1":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=nounits_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_RefSB_b2":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=nounits_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_RefSB_b3":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=nounits_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_RefSB_b4":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=nounits_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_Emissive_b24":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=valid_range_none_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_Emissive_b25":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=radunits_attrs,
                dims=("_rows", "_cols")),
        f"{key_prefix}EV_250_Emissive":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=nounits_attrs,
                dims=("_rows", "_cols")),
    }
    return data


def _get_500m_data(num_scans, rows_per_scan, num_cols):
    data = {
        "Data/EV_Reflectance":
            xr.DataArray(
                da.ones((5, num_scans * rows_per_scan, num_cols), chunks=1024,
                        dtype=np.uint16),
                attrs={
                    "Slope": np.array([1.] * 5, dtype=np.float32), "Intercept": np.array([0.] * 5, dtype=np.float32),
                    "FillValue": 65535,
                    "units": "NO",
                    "valid_range": [0, 4095],
                    "long_name": b"500m Earth View Science Data",
                },
                dims=("_ref_bands", "_rows", "_cols")),
        "Data/EV_Emissive":
            xr.DataArray(
                da.ones((3, num_scans * rows_per_scan, num_cols), chunks=1024,
                        dtype=np.uint16),
                attrs={
                    "Slope": np.array([1.] * 3, dtype=np.float32), "Intercept": np.array([0.] * 3, dtype=np.float32),
                    "FillValue": 65535,
                    "units": "mW/ (m2 cm-1 sr)",
                    "valid_range": [0, 25000],
                    "long_name": b"500m Emissive Bands Earth View "
                                 b"Science Data",
                },
                dims=("_ir_bands", "_rows", "_cols")),
    }
    return data


def _get_1km_data(num_scans, rows_per_scan, num_cols, filetype_info):
    is_mersi1 = filetype_info["file_type"].startswith(("fy3a_mersi1", "fy3b_mersi1", "fy3c_mersi1"))
    is_fy3ab_mersi1 = filetype_info["file_type"].startswith(("fy3a_mersi1", "fy3b_mersi1"))

    fill_value_name = "_FillValue" if is_fy3ab_mersi1 else "FillValue"
    key_prefix = "" if is_fy3ab_mersi1 else "Data/"
    radunits = "NO" if is_mersi1 else "mW/ (m2 cm-1 sr)"

    data = {"Data/EV_1KM_LL":
            xr.DataArray(da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs={"Slope": np.array([1.], dtype=np.float32),
                       "Intercept": np.array([0.], dtype=np.float32),
                       "FillValue": 65535,
                       "units": "NO",
                       "valid_range": [0, 4095],
                       "long_name": b"1km Earth View Science Data"},
                dims=("_rows", "_cols")),
            f"{key_prefix}EV_1KM_RefSB":
            xr.DataArray(da.ones((15, num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs={"Slope": np.array([1.] * 15, dtype=np.float32),
                       "Intercept": np.array([0.] * 15, dtype=np.float32),
                       fill_value_name: 65535,
                       "units": "NO",
                       "valid_range": [0, 4095],
                       "long_name": b"1km Earth View Science Data"},
                dims=("_ref_bands", "_rows", "_cols")),
        "Data/EV_1KM_Emissive":
            xr.DataArray(da.ones((4, num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs={"Slope": np.array([1.] * 4, dtype=np.float32),
                       "Intercept": np.array([0.] * 4, dtype=np.float32),
                       "FillValue": 65535,
                       "units": "mW/ (m2 cm-1 sr)",
                       "valid_range": [0, 25000],
                       "long_name": b"1km Emissive Bands Earth View Science Data"},
                dims=("_ir_bands", "_rows", "_cols")),
        f"{key_prefix}EV_250_Aggr.1KM_RefSB":
            xr.DataArray(da.ones((4, num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs={"Slope": np.array([1.] * 4, dtype=np.float32),
                       "Intercept": np.array([0.] * 4, dtype=np.float32),
                       fill_value_name: 65535,
                       "units": "NO",
                       "valid_range": [0, 4095],
                       "long_name": b"250m Reflective Bands Earth View Science Data Aggregated to 1 km"},
                dims=("_ref250_bands", "_rows", "_cols")),
        f"{key_prefix}EV_250_Aggr.1KM_Emissive":
            xr.DataArray(da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs={"Slope": np.array([1.], dtype=np.float32),
                       "Intercept": np.array([0.], dtype=np.float32),
                       fill_value_name: 65535,
                       "units": radunits,
                       "valid_range": [0, 4095],
                       "long_name": b"250m Emissive Bands Earth View Science Data Aggregated to 1 km"},
                dims=("_rows", "_cols")) if is_mersi1 else
                xr.DataArray(da.ones((2, num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                    attrs={"Slope": np.array([1.] * 2, dtype=np.float32),
                           "Intercept": np.array([0.] * 2, dtype=np.float32),
                           "FillValue": 65535,
                           "units": "mW/ (m2 cm-1 sr)",
                           "valid_range": [0, 4095],
                           "long_name": b"250m Emissive Bands Earth View Science Data Aggregated to 1 km"},
                    dims=("_ir250_bands", "_rows", "_cols")),
        f"{key_prefix}SensorZenith":
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.int16),
                    attrs={
                        "Slope": np.array([.01] * 1, dtype=np.float32),
                        "Intercept": np.array([0.] * 1, dtype=np.float32),
                        "units": "degree",
                        "valid_range": [0, 28000],
                    },
                    dims=("_rows", "_cols")),
            }
    return data


def _get_250m_ll_data(num_scans, rows_per_scan, num_cols):
    # Set some default attributes
    def_attrs = {"FillValue": 65535,
                 "valid_range": [0, 4095],
                 "Slope": np.array([1.], dtype=np.float32), "Intercept": np.array([0.], dtype=np.float32),
                 "long_name": b"250m Earth View Science Data",
                 "units": "mW/ (m2 cm-1 sr)",
                 }
    data = {
        "Data/EV_250_Emissive_b6":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=def_attrs,
                dims=("_rows", "_cols")),
        "Data/EV_250_Emissive_b7":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.uint16),
                attrs=def_attrs,
                dims=("_rows", "_cols")),
    }
    return data


def _get_geo_data(num_scans, rows_per_scan, num_cols, prefix):
    geo = {
        prefix + "Longitude":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.float64),
                attrs={
                    "Slope": np.array([1.] * 1, dtype=np.float64), "Intercept": np.array([0.] * 1, dtype=np.float64),
                    "units": "degree",
                    "valid_range": [-90, 90],
                },
                dims=("_rows", "_cols")),
        prefix + "Latitude":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.float64),
                attrs={
                    "Slope": np.array([1.] * 1, dtype=np.float64), "Intercept": np.array([0.] * 1, dtype=np.float64),
                    "units": "degree",
                    "valid_range": [-180, 180],
                },
                dims=("_rows", "_cols")),
        prefix + "SensorZenith":
            xr.DataArray(
                da.ones((num_scans * rows_per_scan, num_cols), chunks=1024, dtype=np.int16),
                attrs={
                    "Slope": np.array([.01] * 1, dtype=np.float32), "Intercept": np.array([0.] * 1, dtype=np.float32),
                    "units": "degree",
                    "valid_range": [0, 28000],
                },
                dims=("_rows", "_cols")),
    }
    return geo


def make_test_data(dims):
    """Make test data."""
    return xr.DataArray(da.from_array(np.ones([dim for dim in dims], dtype=np.float32) * 10, [dim for dim in dims]))


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    num_scans = 2
    num_cols = 2048

    @property
    def _rows_per_scan(self):
        return self.filetype_info.get("rows_per_scan", 10)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        global_attrs = {
            "/attr/Observing Beginning Date": "2019-01-01",
            "/attr/Observing Ending Date": "2019-01-01",
            "/attr/Observing Beginning Time": "18:27:39.720",
            "/attr/Observing Ending Time": "18:38:36.728",
        }
        fy3a_attrs = {
            "/attr/VIR_Cal_Coeff": np.array([0.0, 1.0, 0.0] * 19, dtype=np.float32),
        }
        fy3b_attrs = {
            "/attr/VIS_Cal_Coeff": np.array([0.0, 1.0, 0.0] * 19, dtype=np.float32),
        }
        fy3d_attrs = {
            "/attr/Solar_Irradiance": np.array([1.0] * 19, dtype=np.float32),
        }

        global_attrs, ftype = self._set_sensor_attrs(global_attrs)
        self._add_tbb_coefficients(global_attrs)
        data = self._get_data_file_content()

        test_content = {}
        test_content.update(global_attrs)
        if "fy3a_mersi1" in self.filetype_info["file_type"]:
            test_content.update(data[0])
            test_content.update(data[1])
        else:
            test_content.update(data)
        if "fy3a_mersi1" in self.filetype_info["file_type"]:
            test_content.update(fy3a_attrs)
        elif "fy3b_mersi1" in self.filetype_info["file_type"]:
            test_content.update(fy3b_attrs)
        elif "mersi2" in self.filetype_info["file_type"]:
            test_content.update(fy3d_attrs)
        if not self.filetype_info["file_type"].startswith(("fy3a_mersi1", "fy3b_mersi1")):
            test_content.update(_get_calibration(self.num_scans, ftype))
        return test_content

    def _set_sensor_attrs(self, global_attrs):
        if "fy3a_mersi1" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3A"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI"
            ftype = "VIS"
        elif "fy3b_mersi1" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3B"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI"
            ftype = "VIS"
        elif "fy3c_mersi1" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3C"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI"
            ftype = "VIS"
        elif "mersi2_l1b" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3D"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI"
            ftype = "VIS"
        elif "mersi_ll" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3E"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI LL"
            ftype = "LL"
        elif "mersi_rm" in self.filetype_info["file_type"]:
            global_attrs["/attr/Satellite Name"] = "FY-3G"
            global_attrs["/attr/Sensor Identification Code"] = "MERSI RM"
            ftype = "RSB"
        return global_attrs, ftype

    def _get_data_file_content(self):
        if "fy3a_mersi1" in self.filetype_info["file_type"]:
            return self._add_band_data_file_content(), self._add_geo_data_file_content()
        else:
            if "_geo" in self.filetype_info["file_type"]:
                return self._add_geo_data_file_content()
            else:
                return self._add_band_data_file_content()

    def _add_geo_data_file_content(self):
        num_scans = self.num_scans
        rows_per_scan = self._rows_per_scan
        return _get_geo_data(num_scans, rows_per_scan,
                             self._num_cols_for_file_type,
                             self._geo_prefix_for_file_type)

    def _add_band_data_file_content(self):
        num_cols = self._num_cols_for_file_type
        num_scans = self.num_scans
        rows_per_scan = self._rows_per_scan
        is_mersill = self.filetype_info["file_type"].startswith("mersi_ll")
        is_1km = "_1000" in self.filetype_info["file_type"]
        is_250m = "_250" in self.filetype_info["file_type"]

        if is_1km:
            return _get_1km_data(num_scans, rows_per_scan, num_cols, self.filetype_info)
        elif is_250m:
            if is_mersill:
                return _get_250m_ll_data(num_scans, rows_per_scan, num_cols)
            else:
                return _get_250m_data(num_scans, rows_per_scan, num_cols, self.filetype_info)
        else:
            return _get_500m_data(num_scans, rows_per_scan, num_cols)

    def _add_tbb_coefficients(self, global_attrs):
        if not self.filetype_info["file_type"].startswith("mersi2_"):
            return

        if "_1000" in self.filetype_info["file_type"]:
            global_attrs["/attr/TBB_Trans_Coefficient_A"] = np.array([1.0] * 6, dtype=np.float32)
            global_attrs["/attr/TBB_Trans_Coefficient_B"] = np.array([0.0] * 6, dtype=np.float32)
        else:
            global_attrs["/attr/TBB_Trans_Coefficient_A"] = np.array([0.0] * 6, dtype=np.float32)
            global_attrs["/attr/TBB_Trans_Coefficient_B"] = np.array([0.0] * 6, dtype=np.float32)

    @property
    def _num_cols_for_file_type(self):
        return self.num_cols if "1000" in self.filetype_info["file_type"] else self.num_cols * 2

    @property
    def _geo_prefix_for_file_type(self):
        if self.filetype_info["file_type"].startswith(("fy3a_mersi1", "fy3b_mersi1")):
            return ""
        else:
            if "1000" in self.filetype_info["file_type"]:
                return "Geolocation/"
            elif "500" in self.filetype_info["file_type"]:
                return "Geolocation/"
            else:
                return ""


def _assert_bands_mda_as_exp(res, band_list, exp_result):
    """Remove test code duplication."""
    exp_cal = exp_result[0]
    exp_unit = exp_result[1]
    exp_shape = exp_result[2]
    for band in band_list:
        assert res[band].attrs["calibration"] == exp_cal
        assert res[band].attrs["units"] == exp_unit
        assert res[band].shape == exp_shape


def _test_find_files_and_readers(reader_config, filenames):
    """Test file and reader search."""
    from satpy.readers.core.loading import load_reader
    reader = load_reader(reader_config)
    files = reader.select_files_from_pathnames(filenames)
    # Make sure we have some files
    reader.create_filehandlers(files)
    assert len(files) == len(filenames)
    assert reader.file_handlers
    return reader


def _test_multi_resolutions(available_datasets, band_list, test_resolution, cal_results_number):
    """Test some bands have multiple resolutions."""
    for band_name in band_list:
        from satpy.dataset.data_dict import get_key
        from satpy.tests.utils import make_dataid
        ds_id = make_dataid(name=band_name, resolution=250)
        if test_resolution == "1000":
            with pytest.raises(KeyError):
                get_key(ds_id, available_datasets, num_results=cal_results_number, best=False)
        else:

            res = get_key(ds_id, available_datasets, num_results=cal_results_number, best=False)
            assert len(res) == cal_results_number

        ds_id = make_dataid(name=band_name, resolution=1000)
        if test_resolution == "250":
            with pytest.raises(KeyError):
                get_key(ds_id, available_datasets, num_results=cal_results_number, best=False)
        else:

            res = get_key(ds_id, available_datasets, num_results=cal_results_number, best=False)
            assert len(res) == cal_results_number


class MERSIL1BTester:
    """Test MERSI1/2/LL/RM L1B Reader."""

    def setup_method(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.mersi_l1b import MERSIL1B
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(MERSIL1B, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()


class MERSI12llL1BTester(MERSIL1BTester):
    """Test MERSI1/2/LL L1B Reader."""

    yaml_file: str = ""
    filenames_1000m: list= []
    filenames_250m: list = []
    filenames_all: list = []
    vis_250_bands: list = []
    ir_250_bands: list = []
    vis_1000_bands: list = []
    ir_1000_bands: list = []
    bands_1000: list = []
    bands_250: list = []

    def test_all_resolutions(self):
        """Test loading data when all resolutions or specific one are available."""
        resolution_list = ["all", "250", "1000"]
        file_list = [self.filenames_all, self.filenames_250m, self.filenames_1000m]

        for resolution in resolution_list:
            filenames = file_list[resolution_list.index(resolution)]
            reader = _test_find_files_and_readers(self.reader_configs, filenames)

            # Verify that we have multiple resolutions for:
            # ---------MERSI-1---------
            #     - Bands 1-4 (visible)
            #     - Bands 5 (IR)
            # ---------MERSI-2---------
            #     - Bands 1-4 (visible)
            #     - Bands 24-25 (IR)
            # ---------MERSI-LL---------
            #     - Bands 6-7 (IR)
            available_datasets = reader.available_dataset_ids
            # Only MERSI-2/LL VIS has radiance calibration
            vis_num_results = 3 if self.yaml_file in ["mersi2_l1b.yaml", "mersi_ll_l1b.yaml"] else 2
            ir_num_results = 3
            _test_multi_resolutions(available_datasets, self.vis_250_bands, resolution, vis_num_results)
            _test_multi_resolutions(available_datasets, self.ir_250_bands, resolution, ir_num_results)

            res = reader.load(self.bands_1000 + self.bands_250)
            for i in res:
                assert res[i].dtype == np.float32
                assert res[i].values.dtype == np.float32

            if resolution != "250":
                assert len(res) == len(self.bands_1000 + self.bands_250)
            else:
                assert len(res) == len(self.bands_250)
                for band in self.bands_1000:
                    with pytest.raises(KeyError):
                        res.__getitem__(band)

            if resolution in ["all", "250"]:
                _assert_bands_mda_as_exp(res, self.vis_250_bands, ("reflectance", "%", (2 * 40, 2048 * 2)))
                _assert_bands_mda_as_exp(res, self.ir_250_bands, ("brightness_temperature", "K", (2 * 40, 2048 * 2)))

                if resolution == "all":
                    _assert_bands_mda_as_exp(res, self.vis_1000_bands, ("reflectance", "%", (2 * 10, 2048)))
                    _assert_bands_mda_as_exp(res, self.ir_1000_bands, ("brightness_temperature", "K", (2 * 10, 2048)))
            else:
                _assert_bands_mda_as_exp(res, self.vis_250_bands, ("reflectance", "%", (2 * 10, 2048)))
                _assert_bands_mda_as_exp(res, self.vis_1000_bands, ("reflectance", "%", (2 * 10, 2048)))
                _assert_bands_mda_as_exp(res, self.ir_250_bands, ("brightness_temperature", "K", (2 * 10, 2048)))
                _assert_bands_mda_as_exp(res, self.ir_1000_bands, ("brightness_temperature", "K", (2 * 10, 2048)))

    def test_counts_calib(self):
        """Test loading data at counts calibration."""
        from satpy.tests.utils import make_dataid
        filenames = self.filenames_all
        reader = _test_find_files_and_readers(self.reader_configs, filenames)

        ds_ids = []
        for band_name in self.bands_1000 + self.bands_250:
            ds_ids.append(make_dataid(name=band_name, calibration="counts"))
        ds_ids.append(make_dataid(name="satellite_zenith_angle"))
        res = reader.load(ds_ids)
        assert len(res) == len(self.bands_1000) + len(self.bands_250) + 1
        _assert_bands_mda_as_exp(res, self.bands_250, ("counts", "1", (2 * 40, 2048 * 2)))
        _assert_bands_mda_as_exp(res, self.bands_1000, ("counts", "1", (2 * 10, 2048)))

    def test_rad_calib(self):
        """Test loading data at radiance calibration. For MERSI-2/LL VIS/IR and MERSI-1 IR."""
        from satpy.tests.utils import make_dataid
        filenames = self.filenames_all
        reader = _test_find_files_and_readers(self.reader_configs, filenames)

        ds_ids = []
        test_bands = self.bands_1000 + self.bands_250 if self.yaml_file in ["mersi2_l1b.yaml", "mersi_ll_l1b.yaml"] \
            else self.ir_250_bands + self.ir_1000_bands

        for band_name in test_bands:
            ds_ids.append(make_dataid(name=band_name, calibration="radiance"))
        res = reader.load(ds_ids)
        assert len(res) == len(test_bands)
        if self.yaml_file in ["mersi2_l1b.yaml", "mersi_ll_l1b.yaml"]:
            _assert_bands_mda_as_exp(res, self.bands_250, ("radiance", "mW/ (m2 cm-1 sr)", (2 * 40, 2048 * 2)))
            _assert_bands_mda_as_exp(res, self.bands_1000, ("radiance", "mW/ (m2 cm-1 sr)", (2 * 10, 2048)))
        else:
            _assert_bands_mda_as_exp(res, self.ir_250_bands, ("radiance", "mW/ (m2 cm-1 sr)", (2 * 40, 2048 * 2)))
            _assert_bands_mda_as_exp(res, self.ir_1000_bands, ("radiance", "mW/ (m2 cm-1 sr)", (2 * 10, 2048)))


class TestFY3AMERSI1L1B(MERSI12llL1BTester):
    """Test the FY3A MERSI1 L1B reader."""

    yaml_file = "fy3a_mersi1_l1b.yaml"
    filenames_1000m = ["FY3A_MERSI_GBAL_L1_20090601_1200_1000M_MS.hdf"]
    filenames_250m = ["FY3A_MERSI_GBAL_L1_20090601_1200_0250M_MS.hdf"]
    filenames_all = filenames_1000m + filenames_250m
    vis_250_bands = ["1", "2", "3", "4"]
    ir_250_bands = ["5"]
    vis_1000_bands = ["6", "7", "8", "11", "15", "19", "20"]
    ir_1000_bands = []
    bands_1000 = vis_1000_bands + ir_1000_bands
    bands_250 = vis_250_bands + ir_250_bands


class TestFY3BMERSI1L1B(MERSI12llL1BTester):
    """Test the FY3B MERSI1 L1B reader."""

    yaml_file = "fy3b_mersi1_l1b.yaml"
    filenames_1000m = ["FY3B_MERSI_GBAL_L1_20110824_1850_1000M_MS.hdf"]
    filenames_250m = ["FY3B_MERSI_GBAL_L1_20110824_1850_0250M_MS.hdf", "FY3B_MERSI_GBAL_L1_20110824_1850_GEOXX_MS.hdf"]
    filenames_all = filenames_1000m + filenames_250m
    vis_250_bands = ["1", "2", "3", "4"]
    ir_250_bands = ["5"]
    vis_1000_bands = ["6", "7", "8", "11", "15", "19", "20"]
    ir_1000_bands = []
    bands_1000 = vis_1000_bands + ir_1000_bands
    bands_250 = vis_250_bands + ir_250_bands


class TestFY3CMERSI1L1B(MERSI12llL1BTester):
    """Test the FY3C MERSI1 L1B reader."""

    yaml_file = "fy3c_mersi1_l1b.yaml"
    filenames_1000m = ["FY3C_MERSI_GBAL_L1_20131002_1835_1000M_MS.hdf", "FY3C_MERSI_GBAL_L1_20131002_1835_GEO1K_MS.hdf"]
    filenames_250m = ["FY3C_MERSI_GBAL_L1_20131002_1835_0250M_MS.hdf", "FY3C_MERSI_GBAL_L1_20131002_1835_GEOQK_MS.hdf"]
    filenames_all = filenames_1000m + filenames_250m
    vis_250_bands = ["1", "2", "3", "4"]
    ir_250_bands = ["5"]
    vis_1000_bands = ["6", "7", "8", "11", "15", "19", "20"]
    ir_1000_bands = []
    bands_1000 = vis_1000_bands + ir_1000_bands
    bands_250 = vis_250_bands + ir_250_bands


class TestFY3DMERSI2L1B(MERSI12llL1BTester):
    """Test the FY3D MERSI2 L1B reader."""

    yaml_file = "mersi2_l1b.yaml"
    filenames_1000m = ["tf2019071182739.FY3D-X_MERSI_1000M_L1B.HDF", "tf2019071182739.FY3D-X_MERSI_GEO1K_L1B.HDF"]
    filenames_250m = ["tf2019071182739.FY3D-X_MERSI_0250M_L1B.HDF", "tf2019071182739.FY3D-X_MERSI_GEOQK_L1B.HDF"]
    filenames_all = filenames_1000m + filenames_250m
    vis_250_bands = ["1", "2", "3", "4"]
    ir_250_bands = ["24", "25"]
    vis_1000_bands = ["5", "8", "9", "11", "15", "17", "19"]
    ir_1000_bands = ["20", "21", "23"]
    bands_1000 = vis_1000_bands + ir_1000_bands
    bands_250 = vis_250_bands + ir_250_bands


class TestFY3EMERSIllL1B(MERSI12llL1BTester):
    """Test the FY3D MERSI2 L1B reader."""

    yaml_file = "mersi_ll_l1b.yaml"
    filenames_1000m = ["FY3E_MERSI_GRAN_L1_20230410_1910_1000M_V0.HDF", "FY3E_MERSI_GRAN_L1_20230410_1910_GEO1K_V0.HDF"]
    filenames_250m = ["FY3E_MERSI_GRAN_L1_20230410_1910_0250M_V0.HDF", "FY3E_MERSI_GRAN_L1_20230410_1910_GEOQK_V0.HDF"]
    filenames_all = filenames_1000m + filenames_250m
    vis_250_bands = []
    ir_250_bands = ["6", "7"]
    vis_1000_bands = ["1"]
    ir_1000_bands = ["2", "3", "5"]
    bands_1000 = vis_1000_bands + ir_1000_bands
    bands_250 = vis_250_bands + ir_250_bands


class TestMERSIRML1B(MERSIL1BTester):
    """Test the FY3E MERSI-RM L1B reader."""

    yaml_file = "mersi_rm_l1b.yaml"
    filenames_500m = ["FY3G_MERSI_GRAN_L1_20230410_1910_0500M_V1.HDF",
                      "FY3G_MERSI_GRAN_L1_20230410_1910_GEOHK_V1.HDF",
                      ]

    def test_500m_resolution(self):
        """Test loading data when all resolutions are available."""
        from satpy.readers.core.loading import load_reader
        filenames = self.filenames_500m
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 2 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers

        res = reader.load(["1", "2", "4", "7"])
        assert len(res) == 4
        assert res["4"].shape == (2 * 10, 4096)
        assert res["1"].attrs["calibration"] == "reflectance"
        assert res["1"].attrs["units"] == "%"
        assert res["2"].shape == (2 * 10, 4096)
        assert res["2"].attrs["calibration"] == "reflectance"
        assert res["2"].attrs["units"] == "%"
        assert res["7"].shape == (20, 2048 * 2)
        assert res["7"].attrs["calibration"] == "brightness_temperature"
        assert res["7"].attrs["units"] == "K"

    def test_rad_calib(self):
        """Test loading data at radiance calibration."""
        from satpy.readers.core.loading import load_reader
        from satpy.tests.utils import make_dataid
        filenames = self.filenames_500m
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 2 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers

        band_names = ["1", "3", "4", "6", "7"]
        ds_ids = []
        for band_name in band_names:
            ds_ids.append(make_dataid(name=band_name, calibration="radiance"))
        res = reader.load(ds_ids)
        assert len(res) == 5
        for band_name in band_names:
            assert res[band_name].shape == (20, 4096)
            assert res[band_name].attrs["calibration"] == "radiance"
            assert res[band_name].attrs["units"] == "mW/ (m2 cm-1 sr)"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for GLM compositors."""


class TestGLMComposites:
    """Test GLM-specific composites."""

    def test_load_composite_yaml(self):
        """Test loading the yaml for this sensor."""
        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        load_compositor_configs_for_sensors(["glm"])

    def test_highlight_compositor(self):
        """Test creating a highlight composite."""
        import dask.array as da
        import numpy as np
        import xarray as xr
        from pyresample.geometry import AreaDefinition

        from satpy.composites.glm import HighlightCompositor
        rows = 5
        cols = 10
        area = AreaDefinition(
            "test", "test", "test",
            {"proj": "eqc", "lon_0": 0.0,
             "lat_0": 0.0},
            cols, rows,
            (-20037508.34, -10018754.17, 20037508.34, 10018754.17))

        comp = HighlightCompositor(
            "c14_highlight",
            prerequisites=("flash_extent_density", "C14"),
            min_hightlight=0.0,
            max_hightlight=1.0,
        )
        flash_extent_density = xr.DataArray(
            da.zeros((rows, cols), chunks=25) + 0.5,
            dims=("y", "x"),
            attrs={"name": "flash_extent_density", "area": area})
        c14_data = np.repeat(np.arange(cols, dtype=np.float64)[None, :], rows, axis=0)
        c14 = xr.DataArray(da.from_array(c14_data, chunks=25) + 303.15,
                           dims=("y", "x"),
                           attrs={
                               "name": "C14",
                               "area": area,
                               "standard_name": "toa_brightness_temperature",
                           })
        res = comp((flash_extent_density, c14))
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "c14_highlight"
        data = res.compute()
        np.testing.assert_almost_equal(data.values.min(), -0.04)
        np.testing.assert_almost_equal(data.values.max(), 1.04)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

"""Tests for the angles in modifiers."""

import contextlib
import datetime as dt
import warnings
from copy import deepcopy
from glob import glob
from typing import Optional, Union
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition, StackedAreaDefinition

import satpy
from satpy.utils import PerformanceWarning

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


def _angle_cache_area_def():
    area = AreaDefinition(
        "test", "", "",
        {"proj": "merc"},
        5, 5,
        (-2500, -2500, 2500, 2500),
    )
    return area


def _angle_cache_stacked_area_def():
    area1 = AreaDefinition(
        "test", "", "",
        {"proj": "merc"},
        5, 2,
        (2500, 500, 7500, 2500),
    )
    area2 = AreaDefinition(
        "test", "", "",
        {"proj": "merc"},
        5, 3,
        (2500, -2500, 7500, 500),
    )
    return StackedAreaDefinition(area1, area2)


def _get_angle_test_data(area_def: Optional[Union[AreaDefinition, StackedAreaDefinition]] = None,
                         chunks: Optional[Union[int, tuple]] = 2,
                         shape: tuple = (5, 5),
                         dims: Optional[tuple] = None,
                         ) -> xr.DataArray:
    if area_def is None:
        area_def = _angle_cache_area_def()
    orb_params = {
        "satellite_nominal_altitude": 12345678,
        "satellite_nominal_longitude": 10.0,
        "satellite_nominal_latitude": 0.0,
    }
    stime = dt.datetime(2020, 1, 1, 12, 0, 0)
    data = da.zeros(shape, chunks=chunks)
    vis = xr.DataArray(data,
                       dims=dims,
                       attrs={
                           "area": area_def,
                           "start_time": stime,
                           "orbital_parameters": orb_params,
                       })
    return vis


def _get_stacked_angle_test_data():
    return _get_angle_test_data(area_def=_angle_cache_stacked_area_def(),
                                chunks=(5, (2, 2, 1)))


def _get_angle_test_data_rgb():
    return _get_angle_test_data(shape=(5, 5, 3), chunks=((2, 2, 1), (2, 2, 1), (1, 1, 1)),
                                dims=("y", "x", "bands"))


def _get_angle_test_data_rgb_nodims():
    return _get_angle_test_data(shape=(3, 5, 5), chunks=((1, 1, 1), (2, 2, 1), (2, 2, 1)))


def _get_angle_test_data_odd_chunks():
    return _get_angle_test_data(chunks=((2, 1, 2), (1, 1, 2, 1)))


def _get_angle_test_data_odd_chunks2():
    return _get_angle_test_data(chunks=((1, 4), (2, 3)))


def _similar_sat_pos_datetime(orig_data, lon_offset=0.04):
    # change data slightly
    new_data = orig_data.copy()
    old_lon = new_data.attrs["orbital_parameters"]["satellite_nominal_longitude"]
    new_data.attrs["orbital_parameters"]["satellite_nominal_longitude"] = old_lon + lon_offset
    new_data.attrs["start_time"] = new_data.attrs["start_time"] + dt.timedelta(hours=36)
    return new_data


def _diff_sat_pos_datetime(orig_data):
    return _similar_sat_pos_datetime(orig_data, lon_offset=0.05)


def _glob_reversed(pat):
    """Behave like glob but force results to be in the wrong order."""
    return sorted(glob(pat), reverse=True)


@contextlib.contextmanager
def _mock_glob_if(mock_glob):
    if mock_glob:
        with mock.patch("satpy.modifiers.angles.glob", _glob_reversed):
            yield
    else:
        yield


def _assert_allclose_if(expect_equal, arr1, arr2):
    if not expect_equal:
        pytest.raises(AssertionError, np.testing.assert_allclose, arr1, arr2)
    else:
        np.testing.assert_allclose(arr1, arr2)


class TestAngleGeneration:
    """Test the angle generation utility functions."""

    @pytest.mark.parametrize(
        ("input_func", "exp_calls"),
        [
            (_get_angle_test_data, 9),
            (_get_stacked_angle_test_data, 3),
            (_get_angle_test_data_rgb, 9),
            (_get_angle_test_data_rgb_nodims, 9),
        ],
    )
    def test_get_angles(self, input_func, exp_calls):
        """Test sun and satellite angle calculation."""
        from satpy.modifiers.angles import get_angles
        data = input_func()

        from pyorbital.orbital import get_observer_look
        with mock.patch("satpy.modifiers.angles.get_observer_look", wraps=get_observer_look) as gol:
            angles = get_angles(data)
            assert all(isinstance(x, xr.DataArray) for x in angles)
            da.compute(*tuple(x.data for x in angles))

        # get_observer_look should have been called once per array chunk
        assert gol.call_count == exp_calls
        # Check arguments of get_orbserver_look() call, especially the altitude
        # unit conversion from meters to kilometers
        args = gol.call_args[0]
        assert args[:4] == (10.0, 0.0, 12345.678, data.attrs["start_time"])

    @pytest.mark.parametrize("forced_preference", ["actual", "nadir"])
    def test_get_angles_satpos_preference(self, forced_preference):
        """Test that 'actual' satellite position is used for generating sensor angles."""
        from satpy.modifiers.angles import get_angles

        input_data1 = _get_angle_test_data()
        # add additional satellite position metadata
        input_data1.attrs["orbital_parameters"]["nadir_longitude"] = 9.0
        input_data1.attrs["orbital_parameters"]["nadir_latitude"] = 0.01
        input_data1.attrs["orbital_parameters"]["satellite_actual_longitude"] = 9.5
        input_data1.attrs["orbital_parameters"]["satellite_actual_latitude"] = 0.005
        input_data1.attrs["orbital_parameters"]["satellite_actual_altitude"] = 12345679
        input_data2 = input_data1.copy(deep=True)
        input_data2.attrs = deepcopy(input_data1.attrs)
        input_data2.attrs["orbital_parameters"]["nadir_longitude"] = 9.1
        input_data2.attrs["orbital_parameters"]["nadir_latitude"] = 0.02
        input_data2.attrs["orbital_parameters"]["satellite_actual_longitude"] = 9.5
        input_data2.attrs["orbital_parameters"]["satellite_actual_latitude"] = 0.005
        input_data2.attrs["orbital_parameters"]["satellite_actual_altitude"] = 12345679

        from pyorbital.orbital import get_observer_look
        with mock.patch("satpy.modifiers.angles.get_observer_look", wraps=get_observer_look) as gol, \
                satpy.config.set(sensor_angles_position_preference=forced_preference):
            angles1 = get_angles(input_data1)
            da.compute(*tuple(x.data for x in angles1))
            angles2 = get_angles(input_data2)
            da.compute(*tuple(x.data for x in angles2))

        # get_observer_look should have been called once per array chunk
        assert gol.call_count == input_data1.data.blocks.size * 2
        if forced_preference == "actual":
            exp_call = mock.call(9.5, 0.005, 12345.679, input_data1.attrs["start_time"], mock.ANY, mock.ANY, 0)
            all_same_calls = [exp_call] * gol.call_count
            gol.assert_has_calls(all_same_calls)
            # the dask arrays should have the same name to prove they are the same computation
            for angle_arr1, angle_arr2 in zip(angles1, angles2):
                assert angle_arr1.data.name == angle_arr2.data.name
        else:
            # nadir 1
            gol.assert_any_call(9.0, 0.01, 12345.679, input_data1.attrs["start_time"], mock.ANY, mock.ANY, 0)
            # nadir 2
            gol.assert_any_call(9.1, 0.02, 12345.679, input_data1.attrs["start_time"], mock.ANY, mock.ANY, 0)

    @pytest.mark.parametrize("force_bad_glob", [False, True])
    @pytest.mark.parametrize(
        ("input2_func", "exp_equal_sun", "exp_num_zarr"),
        [
            (lambda x: x, True, 4),
            (_similar_sat_pos_datetime, False, 4),
            (_diff_sat_pos_datetime, False, 6),
        ]
    )
    @pytest.mark.parametrize(
        ("input_func", "num_normalized_chunks", "exp_zarr_chunks"),
        [
            (_get_angle_test_data, 9, ((2, 2, 1), (2, 2, 1))),
            (_get_stacked_angle_test_data, 3, ((5,), (2, 2, 1))),
            (_get_angle_test_data_odd_chunks, 9, ((2, 1, 2), (1, 1, 2, 1))),
            (_get_angle_test_data_odd_chunks2, 4, ((1, 4), (2, 3))),
            (_get_angle_test_data_rgb, 9, ((2, 2, 1), (2, 2, 1))),
            (_get_angle_test_data_rgb_nodims, 9, ((2, 2, 1), (2, 2, 1))),
        ])
    def test_cache_get_angles(
            self,
            input_func, num_normalized_chunks, exp_zarr_chunks,
            input2_func, exp_equal_sun, exp_num_zarr,
            force_bad_glob, tmp_path):
        """Test get_angles when caching is enabled."""
        from satpy.modifiers.angles import STATIC_EARTH_INERTIAL_DATETIME, get_angles

        # Patch methods
        data = input_func()
        additional_cache = exp_num_zarr > 4

        # Compute angles
        from pyorbital.orbital import get_observer_look
        with mock.patch("satpy.modifiers.angles.get_observer_look", wraps=get_observer_look) as gol, \
                satpy.config.set(cache_lonlats=True, cache_sensor_angles=True, cache_dir=str(tmp_path)), \
                warnings.catch_warnings(record=True) as caught_warnings:
            res = get_angles(data)
            self._check_cached_result(res, exp_zarr_chunks)

            # call again, should be cached
            new_data = input2_func(data)
            with _mock_glob_if(force_bad_glob):
                res2 = get_angles(new_data)
            self._check_cached_result(res2, exp_zarr_chunks)

            res_numpy, res2_numpy = da.compute(tuple(x.data for x in res), tuple(x.data for x in res2))
            for r1, r2 in zip(res_numpy[:2], res2_numpy[:2]):
                _assert_allclose_if(not additional_cache, r1, r2)
            for r1, r2 in zip(res_numpy[2:], res2_numpy[2:]):
                _assert_allclose_if(exp_equal_sun, r1, r2)

            self._check_cache_and_clear(tmp_path, exp_num_zarr)

        if "odd_chunks" in input_func.__name__:
            assert any(w.category is PerformanceWarning for w in caught_warnings)
        else:
            assert not any(w.category is PerformanceWarning for w in caught_warnings)
        assert gol.call_count == num_normalized_chunks * (int(additional_cache) + 1)
        args = gol.call_args_list[0][0]
        assert args[:4] == (10.0, 0.0, 12345.678, STATIC_EARTH_INERTIAL_DATETIME)
        exp_sat_lon = 10.1 if additional_cache else 10.0
        args = gol.call_args_list[-1][0]
        assert args[:4] == (exp_sat_lon, 0.0, 12345.678, STATIC_EARTH_INERTIAL_DATETIME)

    @staticmethod
    def _check_cached_result(results, exp_zarr_chunks):
        assert all(isinstance(x, xr.DataArray) for x in results)
        # output chunks should be consistent
        for angle_data_arr in results:
            assert angle_data_arr.chunks == exp_zarr_chunks

    @staticmethod
    def _check_cache_and_clear(tmp_path, exp_num_zarr):
        from satpy.modifiers.angles import _get_sensor_angles_from_sat_pos, _get_valid_lonlats
        zarr_dirs = glob(str(tmp_path / "*.zarr"))
        assert len(zarr_dirs) == exp_num_zarr  # two for lon/lat, one for sata, one for satz

        _get_valid_lonlats.cache_clear()
        _get_sensor_angles_from_sat_pos.cache_clear()
        zarr_dirs = glob(str(tmp_path / "*.zarr"))
        assert len(zarr_dirs) == 0

    def test_cached_no_chunks_fails(self, tmp_path):
        """Test that trying to pass non-dask arrays and no chunks fails."""
        from satpy.modifiers.angles import _sanitize_args_with_chunks, cache_to_zarr_if

        @cache_to_zarr_if("cache_lonlats", sanitize_args_func=_sanitize_args_with_chunks)
        def _fake_func(data, tuple_arg, chunks):
            return da.from_array(data)

        data = list(range(5))
        with pytest.raises(RuntimeError), \
                satpy.config.set(cache_lonlats=True, cache_dir=str(tmp_path)):
            _fake_func(data, (1, 2, 3), 5)

    def test_cached_result_numpy_fails(self, tmp_path):
        """Test that trying to cache with non-dask arrays fails."""
        from satpy.modifiers.angles import _sanitize_args_with_chunks, cache_to_zarr_if

        @cache_to_zarr_if("cache_lonlats", sanitize_args_func=_sanitize_args_with_chunks)
        def _fake_func(shape, chunks):
            return np.zeros(shape)

        with pytest.raises(ValueError, match="Zarr caching currently only supports dask arrays. Got .*"), \
                satpy.config.set(cache_lonlats=True, cache_dir=str(tmp_path)):
            _fake_func((5, 5), ((5,), (5,)))

    def test_caching_with_array_in_args_warns(self, tmp_path):
        """Test that trying to cache with non-dask arrays fails."""
        from satpy.modifiers.angles import cache_to_zarr_if

        @cache_to_zarr_if("cache_lonlats")
        def _fake_func(array):
            return array + 1

        with pytest.warns(UserWarning, match=".*unhashable argument.*"), \
                satpy.config.set(cache_lonlats=True, cache_dir=str(tmp_path)):
            _fake_func(da.zeros(100))

    def test_caching_with_array_in_args_does_not_warn_when_caching_is_not_enabled(self, tmp_path, recwarn):
        """Test that trying to cache with non-dask arrays fails."""
        from satpy.modifiers.angles import cache_to_zarr_if

        @cache_to_zarr_if("cache_lonlats")
        def _fake_func(array):
            return array + 1

        with satpy.config.set(cache_lonlats=False, cache_dir=str(tmp_path)):
            _fake_func(da.zeros(100))
        assert len(recwarn) == 0

    def test_no_cache_dir_fails(self, tmp_path):
        """Test that 'cache_dir' not being set fails."""
        from satpy.modifiers.angles import _get_sensor_angles_from_sat_pos, get_angles
        data = _get_angle_test_data()
        with pytest.raises(RuntimeError), \
                satpy.config.set(cache_lonlats=True, cache_sensor_angles=True, cache_dir=None):
            get_angles(data)
        with pytest.raises(RuntimeError), \
                satpy.config.set(cache_lonlats=True, cache_sensor_angles=True, cache_dir=None):
            _get_sensor_angles_from_sat_pos.cache_clear()

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize("use_xarray", [False, True])
    def test_relative_azimuth_calculation(self, use_xarray, dtype):
        """Test relative azimuth calculation."""
        from satpy.modifiers.angles import compute_relative_azimuth

        saa = da.from_array(np.array([-120, 40., 0.04, 179.4, 94.2, 12.1], dtype=dtype), chunks=2)
        vaa = da.from_array(np.array([60., 57.7, 175.1, 234.18, 355.4, 12.1], dtype=dtype), chunks=2)
        if use_xarray:
            saa = xr.DataArray(saa, dims=("y",), attrs={"test1": "same", "test2": "diff1"})
            vaa = xr.DataArray(vaa, dims=("y",), attrs={"test1": "same", "test2": "diff2"})

        expected_raa = np.array([180., 17.7, 175.06, 54.78, 98.8, 0.], dtype=dtype)

        raa = compute_relative_azimuth(vaa, saa)
        assert isinstance(raa, xr.DataArray if use_xarray else da.Array)
        computed_raa = raa.compute()
        assert computed_raa.dtype == raa.dtype
        if use_xarray:
            assert raa.attrs == {}
            assert raa.dims == saa.dims
        np.testing.assert_allclose(expected_raa, computed_raa, rtol=2e-7)
        assert raa.chunks == saa.chunks

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_solazi_correction(self, dtype):
        """Test that solar azimuth angles are corrected into the right range."""
        from satpy.modifiers.angles import _get_sun_azimuth_ndarray

        lats = np.array([-80, 40, 0, 40, 80], dtype=dtype)
        lons = np.array([-80, 40, 0, 40, 80], dtype=dtype)

        date = dt.datetime(2022, 1, 5, 12, 50, 0)

        azi = _get_sun_azimuth_ndarray(lats, lons, date)

        assert np.all(azi > 0)
        assert azi.dtype == dtype



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.tropomi_l2 module."""

import datetime as dt
import itertools
import os
import unittest
from unittest import mock

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler

DEFAULT_FILE_DTYPE = np.float32
DEFAULT_FILE_SHAPE = (721, 1440)
DEFAULT_DATE = dt.datetime(2019, 6, 19, 13, 0)
DEFAULT_LAT = np.linspace(-90, 90, DEFAULT_FILE_SHAPE[0], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_LON = np.linspace(-180, 180, DEFAULT_FILE_SHAPE[1], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_FILE_FLOAT_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                                    dtype=DEFAULT_FILE_DTYPE)
DEFAULT_FILE_DATE_DATA = np.clip(DEFAULT_FILE_FLOAT_DATA, 0, 1049)
DEFAULT_FILE_UBYTE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                                    dtype=np.ubyte)
float_variables = ["tpwGrid", "tpwGridPrior", "tpwGridSubseq", "footGridPrior", "footGridSubseq"]
date_variables = ["timeAwayGridPrior", "timeAwayGridSubseq"]
ubyte_variables = ["satGridPrior", "satGridSubseq"]
file_content_attr = dict()


class FakeNetCDF4FileHandlerMimicLow(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content for lower resolution files."""
        dt_s = filename_info.get("start_time", DEFAULT_DATE)
        dt_e = filename_info.get("end_time", DEFAULT_DATE)

        if filetype_info["file_type"] == "mimicTPW2_comp":
            file_content = {
                "/attr/start_time": dt_s.strftime("%Y%m%d.%H%M%S"),
                "/attr/end_time": dt_e.strftime("%Y%m%d.%H%M%S"),
                "/attr/platform_shortname": "aggregated microwave",
                "/attr/sensor": "mimic",
            }
            file_content["latArr"] = DEFAULT_LAT
            file_content["latArr/shape"] = (DEFAULT_FILE_SHAPE[0],)
            file_content["latArr/attr/units"] = "degress_north"

            file_content["lonArr"] = DEFAULT_LON
            file_content["lonArr/shape"] = (DEFAULT_FILE_SHAPE[1],)
            file_content["lonArr/attr/units"] = "degrees_east"

            file_content["/dimension/lat"] = DEFAULT_FILE_SHAPE[0]
            file_content["/dimension/lon"] = DEFAULT_FILE_SHAPE[1]

            for float_var in float_variables:
                file_content[float_var] = DEFAULT_FILE_FLOAT_DATA.reshape(DEFAULT_FILE_SHAPE)
                file_content["{}/shape".format(float_var)] = DEFAULT_FILE_SHAPE
                file_content_attr[float_var] = {"units": "mm"}
            for date_var in date_variables:
                file_content[date_var] = DEFAULT_FILE_DATE_DATA.reshape(DEFAULT_FILE_SHAPE)
                file_content["{}/shape".format(date_var)] = DEFAULT_FILE_SHAPE
                file_content_attr[date_var] = {"units": "minutes"}
            for ubyte_var in ubyte_variables:
                file_content[ubyte_var] = DEFAULT_FILE_UBYTE_DATA.reshape(DEFAULT_FILE_SHAPE)
                file_content["{}/shape".format(ubyte_var)] = DEFAULT_FILE_SHAPE
                file_content_attr[ubyte_var] = {"source_key": "Key: 0: None, 1: NOAA-N, 2: NOAA-P, 3: Metop-A, \
                                                              4: Metop-B, 5: SNPP, 6: SSMI-17, 7: SSMI-18"}

            # convert to xarrays
            for key, val in file_content.items():
                if key == "lonArr" or key == "latArr":
                    file_content[key] = xr.DataArray(val)
                elif isinstance(val, np.ndarray):
                    if val.ndim > 1:
                        file_content[key] = xr.DataArray(val, dims=("y", "x"), attrs=file_content_attr[key])
                    else:
                        file_content[key] = xr.DataArray(val)
            for key in itertools.chain(float_variables, ubyte_variables):
                file_content[key].attrs["_FillValue"] = -999.0
                file_content[key].attrs["name"] = key
                file_content[key].attrs["file_key"] = key
                file_content[key].attrs["file_type"] = self.filetype_info["file_type"]
        else:
            msg = "Wrong Test Reader for file_type {}".format(filetype_info["file_type"])
            raise AssertionError(msg)

        return file_content


class TestMimicTPW2Reader(unittest.TestCase):
    """Test Mimic Reader."""

    yaml_file = "mimicTPW2_comp.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.mimic_TPW2_nc import MimicTPW2FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(MimicTPW2FileHandler, "__bases__", (FakeNetCDF4FileHandlerMimicLow,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic initialization of this reader."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "comp20190619.130000.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_mimic_float(self):
        """Load TPW mimic float data."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.mimic_TPW2_nc.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "comp20190619.130000.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(float_variables)
        assert len(ds) == len(float_variables)
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "aggregated microwave"
            assert d.attrs["sensor"] == "mimic"
            assert d.attrs["units"] == "mm"
            assert "area" in d.attrs
            assert d.attrs["area"] is not None

    def test_load_mimic_timedelta(self):
        """Load TPW mimic timedelta data (data latency variables)."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.mimic_TPW2_nc.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "comp20190619.130000.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(date_variables)
        assert len(ds) == len(date_variables)
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "aggregated microwave"
            assert d.attrs["sensor"] == "mimic"
            assert d.attrs["units"] == "minutes"
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert d.dtype == DEFAULT_FILE_DTYPE

    def test_load_mimic_ubyte(self):
        """Load TPW mimic sensor grids."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.mimic_TPW2_nc.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "comp20190619.130000.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(ubyte_variables)
        assert len(ds) == len(ubyte_variables)
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "aggregated microwave"
            assert d.attrs["sensor"] == "mimic"
            assert "source_key" in d.attrs
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert d.dtype == np.uint8



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the LI instrument."""

import logging

import numpy as np
import xarray as xr

from satpy.composites.core import CompositeBase

LOG = logging.getLogger(__name__)


class LightningTimeCompositor(CompositeBase):
      """Compositor class for lightning visualisation based on time.

      The compositor normalises the lightning event times between 0 and 1.
      The value 1 corresponds to the latest lightning event and the value 0 corresponds
      to the latest lightning event - time_range. The time_range is defined in the composite recipe
      and is in minutes.
      """
      def __init__(self, name, prerequisites=None, optional_prerequisites=None, **kwargs):
          """Initialisation of the class."""
          super().__init__(name, prerequisites, optional_prerequisites, **kwargs)
          # Get the time_range which is in minute
          self.time_range = self.attrs["time_range"]
          self.standard_name = self.attrs["standard_name"]
          self.reference_time_attr = self.attrs["reference_time"]


      def _normalize_time(self, data: xr.DataArray, attrs: dict) -> xr.DataArray:
          """Normalize the time in the range between [end_time, end_time - time_range].

          The range of the normalised data is between 0 and 1 where 0 corresponds to the date end_time - time_range
          and 1 to the end_time. Where end_times represent the latest lightning event and time_range is the range of
          time in minutes visualised in the composite.
          The dates that are earlier to end_time - time_range are set to NaN.

          Args:
              data: datas containing dates to be normalised
              attrs: Attributes suited to the flash_age composite

          Returns:
              Normalised time
          """
          # Compute the maximum time value
          end_time = np.array(np.datetime64(data.attrs[self.reference_time_attr]))
          # Compute the minimum time value based on the time range
          begin_time = end_time - np.timedelta64(self.time_range, "m")
          # Invalidate values that are before begin_time
          condition_time = data >= begin_time
          data = data.where(condition_time)

          # raise a warning if data is empty after filtering
          if np.all(np.isnan(data)) :
              LOG.warning(f"All the flash_age events happened before {begin_time}, the composite will be empty.")

          # Normalize the time values
          normalized_data = (data - begin_time) / (end_time - begin_time)
          # Ensure the result is still an xarray.DataArray
          return xr.DataArray(normalized_data, dims=data.dims, coords=data.coords, attrs=attrs)


      @staticmethod
      def _update_missing_metadata(existing_attrs, new_attrs):
          for key, val in new_attrs.items():
              if key not in existing_attrs and val is not None:
                  existing_attrs[key] = val

      def _redefine_metadata(self, attrs: dict) -> dict:
          """Modify the standard_name and name metadatas.

          Args:
              attrs: data's attributes

          Returns:
              updated attributes
          """
          attrs["name"] = self.standard_name
          attrs["standard_name"] = self.standard_name
          return attrs


      def __call__(self, projectables, nonprojectables=None, **attrs):
          """Normalise the dates."""
          data = projectables[0]
          new_attrs = data.attrs.copy()
          self._update_missing_metadata(new_attrs, attrs)
          new_attrs = self._redefine_metadata(new_attrs)
          return self._normalize_time(data, new_attrs)



"""Tests for the SGLI L1B backend."""

import datetime as dt
import sys

import dask
import h5py
import numpy as np
import pytest

from satpy.readers.sgli_l1b import HDF5SGLI
from satpy.tests.utils import RANDOM_GEN

START_TIME = dt.datetime.now()
END_TIME = START_TIME + dt.timedelta(minutes=5)
FULL_KM_ARRAY = np.arange(1955 * 1250, dtype=np.uint16).reshape((1955, 1250))
MASK = 16383
LON_LAT_ARRAY = np.arange(197 * 126, dtype=np.float32).reshape((197, 126))
AZI_ARRAY = RANDOM_GEN.integers(-180 * 100, 180 * 100, size=(197, 126), dtype=np.int16)
ZEN_ARRAY = RANDOM_GEN.integers(0, 180 * 100, size=(197, 126), dtype=np.int16)


@pytest.fixture(scope="module")
def sgli_vn_file(tmp_path_factory):
    """Create a stub VN file."""
    filename = tmp_path_factory.mktemp("data") / "test_vn_file.h5"
    with h5py.File(filename, "w") as h5f:
        global_attributes = h5f.create_group("Global_attributes")
        global_attributes.attrs["Scene_start_time"] = np.array([START_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                               dtype="|S21")
        global_attributes.attrs["Scene_end_time"] = np.array([END_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                             dtype="|S21")

        image_data = h5f.create_group("Image_data")
        image_data.attrs["Number_of_lines"] = 1955
        image_data.attrs["Number_of_pixels"] = 1250

        vn01 = image_data.create_dataset("Lt_VN01", data=FULL_KM_ARRAY, chunks=(116, 157))
        vn01.attrs["Slope_reflectance"] = np.array([5e-05], dtype=np.float32)
        vn01.attrs["Offset_reflectance"] = np.array([-0.05], dtype=np.float32)
        vn01.attrs["Slope"] = np.array([0.02], dtype=np.float32)
        vn01.attrs["Offset"] = np.array([-25], dtype=np.float32)
        vn01.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        vn01.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")

        add_downsampled_geometry_data(h5f)

    return filename

@pytest.fixture(scope="module")
def sgli_ir_file(tmp_path_factory):
    """Create a stub IR file."""
    filename = tmp_path_factory.mktemp("data") / "test_ir_file.h5"
    with h5py.File(filename, "w") as h5f:
        global_attributes = h5f.create_group("Global_attributes")
        global_attributes.attrs["Scene_start_time"] = np.array([START_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                               dtype="|S21")
        global_attributes.attrs["Scene_end_time"] = np.array([END_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                             dtype="|S21")

        image_data = h5f.create_group("Image_data")
        image_data.attrs["Number_of_lines"] = 1854
        image_data.attrs["Number_of_pixels"] = 1250

        sw01 = image_data.create_dataset("Lt_SW01", data=FULL_KM_ARRAY, chunks=(116, 157))
        sw01.attrs["Slope_reflectance"] = np.array([5e-05], dtype=np.float32)
        sw01.attrs["Offset_reflectance"] = np.array([0.0], dtype=np.float32)
        sw01.attrs["Slope"] = np.array([0.02], dtype=np.float32)
        sw01.attrs["Offset"] = np.array([-25], dtype=np.float32)
        sw01.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        sw01.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")


        ti01 = image_data.create_dataset("Lt_TI01", data=FULL_KM_ARRAY, chunks=(116, 157))
        ti01.attrs["Slope"] = np.array([0.0012], dtype=np.float32)
        ti01.attrs["Offset"] = np.array([-1.65], dtype=np.float32)
        ti01.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        ti01.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")
        ti01.attrs["Center_wavelength"] = np.array([12000], dtype=np.float32)

        add_downsampled_geometry_data(h5f)

    return filename

@pytest.fixture(scope="module")
def sgli_pol_file(tmp_path_factory):
    """Create a POL stub file."""
    filename = tmp_path_factory.mktemp("data") / "test_pol_file.h5"
    with h5py.File(filename, "w") as h5f:
        global_attributes = h5f.create_group("Global_attributes")
        global_attributes.attrs["Scene_start_time"] = np.array([START_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                               dtype="|S21")
        global_attributes.attrs["Scene_end_time"] = np.array([END_TIME.strftime("%Y%m%d %H:%M:%S.%f")[:-3]],
                                                             dtype="|S21")

        image_data = h5f.create_group("Image_data")
        image_data.attrs["Number_of_lines"] = 1854
        image_data.attrs["Number_of_pixels"] = 1250

        p1_0 = image_data.create_dataset("Lt_P1_0", data=FULL_KM_ARRAY, chunks=(116, 157))
        p1_0.attrs["Slope_reflectance"] = np.array([5e-05], dtype=np.float32)
        p1_0.attrs["Offset_reflectance"] = np.array([0.0], dtype=np.float32)
        p1_0.attrs["Slope"] = np.array([0.02], dtype=np.float32)
        p1_0.attrs["Offset"] = np.array([-25], dtype=np.float32)
        p1_0.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        p1_0.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")


        p1_m60 = image_data.create_dataset("Lt_P1_m60", data=FULL_KM_ARRAY, chunks=(116, 157))
        p1_m60.attrs["Slope_reflectance"] = np.array([5e-05], dtype=np.float32)
        p1_m60.attrs["Offset_reflectance"] = np.array([-60.0], dtype=np.float32)
        p1_m60.attrs["Slope"] = np.array([0.0012], dtype=np.float32)
        p1_m60.attrs["Offset"] = np.array([-1.65], dtype=np.float32)
        p1_m60.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        p1_m60.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")

        p1_60 = image_data.create_dataset("Lt_P1_60", data=FULL_KM_ARRAY, chunks=(116, 157))
        p1_60.attrs["Slope_reflectance"] = np.array([5e-05], dtype=np.float32)
        p1_60.attrs["Offset_reflectance"] = np.array([60.0], dtype=np.float32)
        p1_60.attrs["Slope"] = np.array([0.0012], dtype=np.float32)
        p1_60.attrs["Offset"] = np.array([-1.65], dtype=np.float32)
        p1_60.attrs["Mask"] = np.array([16383], dtype=np.uint16)
        p1_60.attrs["Bit00(LSB)-13"] = np.array([b"Digital Number\n16383 : Missing value\n16382 : Saturation value"],
                                               dtype="|S61")

        geometry_data = h5f.create_group("Geometry_data")
        longitude = geometry_data.create_dataset("Longitude", data=FULL_KM_ARRAY.astype(np.float32), chunks=(47, 63))
        longitude.attrs["Resampling_interval"] = 1
        latitude = geometry_data.create_dataset("Latitude", data=FULL_KM_ARRAY.astype(np.float32), chunks=(47, 63))
        latitude.attrs["Resampling_interval"] = 1

        return filename

def add_downsampled_geometry_data(h5f):
    """Add downsampled geometry data to an h5py file instance."""
    geometry_data = h5f.create_group("Geometry_data")
    longitude = geometry_data.create_dataset("Longitude", data=LON_LAT_ARRAY, chunks=(47, 63))
    longitude.attrs["Resampling_interval"] = 10
    latitude = geometry_data.create_dataset("Latitude", data=LON_LAT_ARRAY, chunks=(47, 63))
    latitude.attrs["Resampling_interval"] = 10

    angles_slope = np.array([0.01], dtype=np.float32)
    angles_offset = np.array([0], dtype=np.float32)

    azimuth = geometry_data.create_dataset("Sensor_azimuth", data=AZI_ARRAY, chunks=(47, 63))
    azimuth.attrs["Resampling_interval"] = 10
    azimuth.attrs["Slope"] = angles_slope
    azimuth.attrs["Offset"] = angles_offset
    zenith = geometry_data.create_dataset("Sensor_zenith", data=ZEN_ARRAY, chunks=(47, 63))
    zenith.attrs["Resampling_interval"] = 10
    zenith.attrs["Slope"] = angles_slope
    zenith.attrs["Offset"] = angles_offset

    sazimuth = geometry_data.create_dataset("Solar_azimuth", data=AZI_ARRAY, chunks=(47, 63))
    sazimuth.attrs["Resampling_interval"] = 10
    sazimuth.attrs["Slope"] = angles_slope
    sazimuth.attrs["Offset"] = angles_offset
    szenith = geometry_data.create_dataset("Solar_zenith", data=ZEN_ARRAY, chunks=(47, 63))
    szenith.attrs["Resampling_interval"] = 10
    szenith.attrs["Slope"] = angles_slope
    szenith.attrs["Offset"] = angles_offset


def test_start_time(sgli_vn_file):
    """Test that the start time is extracted."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    microseconds = START_TIME.microsecond % 1000
    assert handler.start_time == START_TIME - dt.timedelta(microseconds=microseconds)


def test_end_time(sgli_vn_file):
    """Test that the end time is extracted."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    microseconds = END_TIME.microsecond % 1000
    assert handler.end_time == END_TIME - dt.timedelta(microseconds=microseconds)

def test_get_dataset_counts(sgli_vn_file):
    """Test that counts can be extracted from a file."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VN1", resolution=1000, polarization=None, calibration="counts")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "",
                                    "standard_name": ""})
    assert np.allclose(res, FULL_KM_ARRAY & MASK)
    assert res.dtype == np.uint16
    assert res.attrs["platform_name"] == "GCOM-C1"
    assert res.attrs["sensor"] == "sgli"

def test_get_dataset_for_unknown_channel(sgli_vn_file):
    """Test that counts can be extracted from a file."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VIN", resolution=1000, polarization=None, calibration="counts")
    with pytest.raises(KeyError):
        handler.get_dataset(did, {"file_key": "Image_data/Lt_VIN01", "units": "",
                                        "standard_name": ""})

def test_get_vn_dataset_reflectances(sgli_vn_file):
    """Test that the vn datasets can be calibrated to reflectances."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VN1", resolution=1000, polarization=None, calibration="reflectance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "%",
                                    "standard_name": ""})
    assert np.allclose(res[0, :] / 100, FULL_KM_ARRAY[0, :] * 5e-5 - 0.05)
    assert res.dtype == np.float32
    assert res.dims == ("y", "x")
    assert res.attrs["units"] == "%"

def test_get_vn_dataset_radiance(sgli_vn_file):
    """Test that datasets can be calibrated to radiance."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VN1", resolution=1000, polarization=None, calibration="radiance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "W m-2 um-1 sr-1",
                                    "standard_name": "toa_outgoing_radiance_per_unit_wavelength"})
    assert np.allclose(res[0, :], FULL_KM_ARRAY[0, :] * np.float32(0.02) - 25)
    assert res.dtype == np.float32
    assert res.attrs["units"] == "W m-2 um-1 sr-1"
    assert res.attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"

def test_channel_is_masked(sgli_vn_file):
    """Test that channels are masked for no-data."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VN1", resolution=1000, polarization=None, calibration="counts")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "",
                                    "standard_name": ""})
    assert res.max() == MASK

def test_missing_values_are_masked(sgli_vn_file):
    """Check that missing values are masked."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="VN1", resolution=1000, polarization=None, calibration="radiance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "",
                                    "standard_name": ""})
    assert np.isnan(res).sum() == 149

def test_channel_is_chunked(sgli_vn_file):
    """Test that the channel data is chunked."""
    with dask.config.set({"array.chunk-size": "1MiB"}):
        handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
        did = dict(name="VN1", resolution=1000, polarization=None, calibration="counts")
        res = handler.get_dataset(did, {"file_key": "Image_data/Lt_VN01", "units": "",
                                    "standard_name": ""})
        assert res.chunks[0][0] > 116

@pytest.mark.skipif(sys.version_info < (3, 10), reason="Python 3.10 or higher needed for geotiepoints")
def test_loading_lon_lat(sgli_vn_file):
    """Test that loading lons and lats works."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="longitude_v", resolution=1000, polarization=None)
    res = handler.get_dataset(did, {"file_key": "Geometry_data/Longitude", "units": "",
                                    "standard_name": ""})
    assert res.shape == (1955, 1250)
    assert res.chunks is not None
    assert res.dtype == np.float32
    assert res.dims == ("y", "x")

@pytest.mark.skipif(sys.version_info < (3, 10), reason="Python 3.10 or higher needed for geotiepoints")
def test_loading_sensor_angles(sgli_vn_file):
    """Test loading the satellite angles."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="satellite_zenith_angle", resolution=1000, polarization=None)
    res = handler.get_dataset(did, {"file_key": "Geometry_data/Sensor_zenith", "units": "",
                                    "standard_name": ""})
    assert res.shape == (1955, 1250)
    assert res.chunks is not None
    assert res.dtype == np.float32
    assert res.min() >= 0

@pytest.mark.skipif(sys.version_info < (3, 10), reason="Python 3.10 or higher needed for geotiepoints")
def test_loading_solar_angles(sgli_vn_file):
    """Test loading sun angles."""
    handler = HDF5SGLI(sgli_vn_file, {"resolution": "L"}, {})
    did = dict(name="solar_azimuth_angle", resolution=1000, polarization=None)
    res = handler.get_dataset(did, {"file_key": "Geometry_data/Sensor_zenith", "units": "",
                                    "standard_name": ""})
    assert res.shape == (1955, 1250)
    assert res.chunks is not None
    assert res.dtype == np.float32
    assert res.max() <= 180

def test_get_sw_dataset_reflectances(sgli_ir_file):
    """Test getting SW dataset reflectances."""
    handler = HDF5SGLI(sgli_ir_file, {"resolution": "L"}, {})
    did = dict(name="SW1", resolution=1000, polarization=None, calibration="reflectance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_SW01", "units": "",
                                    "standard_name": ""})
    assert np.allclose(res[0, :] / 100, FULL_KM_ARRAY[0, :] * 5e-5)
    assert res.dtype == np.float32

def test_get_ti_dataset_radiance(sgli_ir_file):
    """Test getting thermal IR radiances."""
    handler = HDF5SGLI(sgli_ir_file, {"resolution": "L"}, {})
    did = dict(name="TI1", resolution=1000, polarization=None, calibration="radiance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_TI01", "units": "",
                                    "standard_name": ""})
    assert np.allclose(res[0, :], FULL_KM_ARRAY[0, :] * np.float32(0.0012) - 1.65)
    assert res.dtype == np.float32

def test_get_ti_dataset_bt(sgli_ir_file):
    """Test getting brightness temperatures for IR channels."""
    handler = HDF5SGLI(sgli_ir_file, {"resolution": "L"}, {})
    did = dict(name="TI1", resolution=1000, polarization=None, calibration="brightness_temperature")
    with pytest.raises(NotImplementedError):
        _ = handler.get_dataset(did, {"file_key": "Image_data/Lt_TI01", "units": "K",
                                    "standard_name": "toa_brightness_temperature"})

@pytest.mark.skipif(sys.version_info < (3, 10), reason="Python 3.10 or higher needed for geotiepoints")
def test_get_ti_lon_lats(sgli_ir_file):
    """Test getting the lons and lats for IR channels."""
    handler = HDF5SGLI(sgli_ir_file, {"resolution": "L"}, {})
    did = dict(name="longitude_ir", resolution=1000, polarization=None)
    res = handler.get_dataset(did, {"file_key": "Geometry_data/Longitude", "units": "",
                                    "standard_name": ""})
    assert res.shape == (1854, 1250)
    assert res.chunks is not None
    assert res.dtype == np.float32

@pytest.mark.parametrize("polarization", [0, -60, 60])
def test_get_polarized_dataset_reflectance(sgli_pol_file, polarization):
    """Test getting polarized reflectances."""
    handler = HDF5SGLI(sgli_pol_file, {"resolution": "L"}, {})
    did = dict(name="P1", resolution=1000, polarization=polarization, calibration="reflectance")
    res = handler.get_dataset(did, {"file_key": "Image_data/Lt_P1_{polarization}", "units": "%",
                                    "standard_name": "toa_bidirectional_reflectance"})
    assert res.dtype == np.float32
    expected = (FULL_KM_ARRAY[0, :] * np.float32(5e-5) + np.float32(polarization)) * 100
    np.testing.assert_allclose(res[0, :], expected)
    assert res.attrs["units"] == "%"
    assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"

def test_get_polarized_longitudes(sgli_pol_file):
    """Test getting polarized reflectances."""
    handler = HDF5SGLI(sgli_pol_file, {"resolution": "L"}, {})
    did = dict(name="longitude", resolution=1000, polarization=0)
    res = handler.get_dataset(did, {"file_key": "Geometry_data/Longitude", "units": "",
                                    "standard_name": ""})
    assert res.dtype == np.float32
    expected = FULL_KM_ARRAY.astype(np.float32)
    np.testing.assert_allclose(res, expected)
